diff --git a/CHANGELOG b/CHANGELOG
index c774dd0..12a325f 100644
--- a/CHANGELOG
+++ b/CHANGELOG
@@ -2,6 +2,11 @@
 Changes since U-Boot 1.1.4:
 ======================================================================
 
+* Add port of FAT filesystem driver from the Rockbox (www.rockbox.org)
+  project.  Driver provides a "roughly POSIX-compatible file IO API"
+  including read, write, open, close and ls.  
+  Patch by Keith Outwater, 08 May 2006
+
 * Coding Style cleanup
 
 * Write RTC seconds first to maintain settings integrity per
diff --git a/CREDITS b/CREDITS
index f91fa3e..b76059c 100644
--- a/CREDITS
+++ b/CREDITS
@@ -326,8 +326,9 @@ E: torkun@nextio.com
 D: Support for Cogent CSB272 & CSB472 boards
 
 N: Keith Outwater
-E: keith_outwater@mvis.com
+E: outwater4@comcast.net
 D: Support for generic/custom MPC860T boards (GEN860T, GEN860T_SC)
+D: Port Rockbox (www.rockbox.org) FAT filesystem driver. Add FAT write support.
 
 N: Frank Panno
 E: fpanno@delphintech.com
diff --git a/README b/README
index 3ffef62..f460088 100644
--- a/README
+++ b/README
@@ -746,6 +746,42 @@ The following options need to be configu
 		Note that if the RTC uses I2C, then the I2C interface
 		must also be configured. See I2C Support, below.
 
+ - Rockbox FAT Filesystem Driver:
+ 		U-Boot includes a port of the FAT filesystem driver from the
+ 		Rockbox project (www.rockbox.org).  To configure U-Boot to
+ 		use thus driver (and to NOT use the default FAT12/16/32
+ 		read-only driver), define CFG_ROCKBOX_FAT along with
+ 		CFG_CMD_FAT.
+ 
+ 		Note that the rockbox FAT driver does not support FAT12, but
+ 		it does support FAT16 and FAT32 (with long file names) and you
+ 		get FAT read and write functionality too.
+ 
+ 		CFG_ROCKBOX_FAT_MAX_OPEN_FILES (optional) -  define this to
+ 		specify the maximum number of files that may be open at one
+ 		time. The default is 1.  This should be sufficient unless you
+ 		are planning to use the posix-like interface to a FAT
+ 		filesystem and need more than one file open at a time.
+ 
+ 		CFG_ROCKBOX_FAT_MAX_OPEN_DIRS (optional) - define this to
+ 		specify the maximum number of simultaneously open directories.
+ 		The default is 8.
+ 
+ 		CFG_ROCKBOX_FAT_MAX_SECS_PER_XFER (optional) - define this
+ 		parameter in order to allow the filesystem driver to request
+ 		transfer (read or write) of more than 256 sectors pr transfer
+ 		request.  If you override the default, MAKE SURE that your
+ 		disk driver can handle more than 256 sectors per transfer
+ 		request.
+ 
+ 		The rockbox driver expects to set the 'errno' variable, so
+ 		your BSP needs to declare it as follows:
+ 
+ 			int errno;
+ 
+ 		If you want to include strerror() support, define
+ 		CONFIG_STRERROR.
+ 
 - Timestamp Support:
 
 		When CONFIG_TIMESTAMP is selected, the timestamp
diff --git a/common/cmd_fat.c b/common/cmd_fat.c
index 6844c10..0b3e19d 100644
--- a/common/cmd_fat.c
+++ b/common/cmd_fat.c
@@ -2,6 +2,10 @@
  * (C) Copyright 2002
  * Richard Jones, rjones@nexus-tech.net
  *
+ * Keith Outwater (outwater4@comcast.net) - add additional
+ *   FAT commands for use with rockbox (www.rockbox.org) FAT
+ *   filesystem driver.
+ *
  * See file CREDITS for list of people who contributed to this
  * project.
  *
@@ -85,7 +89,8 @@ int do_fat_fsload (cmd_tbl_t *cmdtp, int
 	char *ep;
 
 	if (argc < 5) {
-		printf ("usage: fatload <interface> <dev[:part]> <addr> <filename> [bytes]\n");
+		printf ("usage: fatload <interface> <dev[:part]> <addr> "
+				"<filename> [bytes]\n");
 		return 1;
 	}
 	dev = (int)simple_strtoul (argv[2], &ep, 16);
@@ -102,7 +107,8 @@ int do_fat_fsload (cmd_tbl_t *cmdtp, int
 		part = (int)simple_strtoul(++ep, NULL, 16);
 	}
 	if (fat_register_device(dev_desc,part)!=0) {
-		printf ("\n** Unable to use %s %d:%d for fatload **\n",argv[1],dev,part);
+		printf ("\n** Unable to use %s %d:%d for fatload **\n",
+				argv[1],dev,part);
 		return 1;
 	}
 	offset = simple_strtoul (argv[3], NULL, 16);
@@ -113,7 +119,8 @@ int do_fat_fsload (cmd_tbl_t *cmdtp, int
 	size = file_fat_read (argv[4], (unsigned char *) offset, count);
 
 	if(size==-1) {
-		printf("\n** Unable to read \"%s\" from %s %d:%d **\n",argv[4],argv[1],dev,part);
+		printf("\n** Unable to read \"%s\" from %s %d:%d **\n",
+			   argv[4],argv[1],dev,part);
 		return 1;
 	}
 
@@ -126,13 +133,402 @@ int do_fat_fsload (cmd_tbl_t *cmdtp, int
 }
 
 
+#ifdef CFG_ROCKBOX_FAT
 U_BOOT_CMD(
 	fatload,	6,	0,	do_fat_fsload,
 	"fatload - load binary file from a dos filesystem\n",
-	"<interface> <dev[:part]>  <addr> <filename> [bytes]\n"
+	"<interface> <dev[:part]> <addr> <filename> [bytes]\n"
 	"    - load binary file 'filename' from 'dev' on 'interface'\n"
 	"      to address 'addr' from dos filesystem\n"
 );
+#else
+U_BOOT_CMD(
+	fatload,	6,	0,	do_fat_fsload,
+	"fatload - load binary file from a dos FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <addr> <filename> [bytes]\n"
+	"    - load binary file 'filename' from 'dev' on 'interface'\n"
+	"      to address 'addr' from dos FAT16/32 filesystem\n"
+);
+#endif
+
+#ifdef CFG_ROCKBOX_FAT
+int
+do_fat_fswrite (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	long size;
+	unsigned long offset;
+	unsigned long count;
+	char buf[12];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 6) {
+		printf ("usage: fatwrite <interface> <dev[:part]> "
+				"<addr> <filename> <bytes>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatwrite **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	offset = simple_strtoul (argv[3], NULL, 16);
+	count = simple_strtoul (argv[5], NULL, 16);
+	size = file_fat_write(argv[4], (unsigned char *)offset, count);
+
+	if (size <= 0) {
+		printf("\n** Unable to write\"%s\" to %s %d:%d **\n",
+			   argv[4],argv[1],dev,part);
+		return 1;
+	}
+
+	printf ("\n%ld bytes written\n", size);
+	sprintf(buf, "%lX", size);
+	setenv("filesize", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatwrite,	6,	0,	do_fat_fswrite,
+	"fatwrite- store a binary file to a dos filesystem\n",
+	"<interface> <dev[:part]> <addr> <filename> <bytes>\n"
+	"    - store binary file 'filename' to 'dev' on 'interface'\n"
+	"      containing a dos filesystem from address 'addr'\n"
+	"      with length 'bytes'\n"
+);
+
+int
+do_fat_fsrm (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	char buf[256];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 4) {
+		printf ("usage: fatrm <interface> <dev[:part]> <filename>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatrm **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	if (file_fat_rm(argv[3]) < 0) {
+		printf("\n** Unable to remove \"%s\" **\n", argv[3]);
+		return 1;
+	}
+
+	strncpy(buf, argv[3], sizeof(buf));
+	setenv("rm_filename", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatrm,	4,	0,	do_fat_fsrm,
+	"fatrm   - remove a file in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <filename>\n"
+	"    - remove 'filename' in the FAT16/32 filesytem\n"
+	"      on 'dev' on 'interface'\n"
+);
+
+int
+do_fat_fsmkdir (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	char buf[256];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 4) {
+		printf ("usage: fatmkdir <interface> <dev[:part]> <dirname>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatmkdir **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	if (file_fat_mkdir(argv[3]) < 0) {
+		printf("\n** Unable to make dir \"%s\" **\n", argv[3]);
+		return 1;
+	}
+
+	strncpy(buf, argv[3], sizeof(buf));
+	setenv("mkdir_dirname", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatmkdir,	4,	0,	do_fat_fsmkdir,
+	"fatmkdir- make a directory in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <dirname>\n"
+	"    - make 'dirname' in the FAT16/32 filesytem on 'dev'\n"
+	"      on 'interface'\n"
+);
+
+int
+do_fat_fsrmdir (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	char buf[256];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 4) {
+		printf ("usage: fatrmdir <interface> <dev[:part]> <dirname>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatrmdir **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	if (file_fat_rmdir(argv[3]) < 0) {
+		printf("\n** Unable to remove dir \"%s\" **\n", argv[3]);
+		return 1;
+	}
+
+	strncpy(buf, argv[3], sizeof(buf));
+	setenv("rm_dirname", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatrmdir,	4,	0,	do_fat_fsrmdir,
+	"fatrmdir- remove an empty directory in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <dirname>\n"
+	"    - remove 'dirname' in the FAT16/32 filesytem on 'dev'\n"
+	"      on 'interface'\n"
+);
+
+int
+do_fat_fscd (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	char buf[256];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 4) {
+		printf ("usage: fatcd <interface> <dev[:part]> <dirname>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatcd **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	if (file_fat_cd(argv[3]) < 0) {
+		printf("\n** Unable to change to dir \"%s\" **\n", argv[3]);
+		return 1;
+	}
+
+	strncpy(buf, argv[3], sizeof(buf));
+	setenv("cwd", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatcd,	4,	0,	do_fat_fscd,
+	"fatcd   - change to the specified dir in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <dirname>\n"
+	"    - change to 'dirname' in the FAT16/32 filesytem on 'dev'\n"
+	"      on 'interface'\n"
+);
+
+int
+do_fat_fspwd (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 3) {
+		printf ("usage: fatpwd <interface> <dev[:part]>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatpwd **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	file_fat_pwd();
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatpwd,	3,	0,	do_fat_fspwd,
+	"fatpwd  - print the current working dir in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]>\n"
+	"    - print the CWD in the FAT16/32 filesytem on 'dev'\n"
+	"      on 'interface'\n"
+);
+
+int
+do_fat_fsmv (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
+{
+	char buf[256];
+	block_dev_desc_t *dev_desc=NULL;
+	int dev=0;
+	int part=1;
+	char *ep;
+
+	if (argc < 5) {
+		printf ("usage: fatmv <interface> <dev[:part]> <oldname> <newname>\n");
+		return 1;
+	}
+
+	dev = (int)simple_strtoul (argv[2], &ep, 16);
+	dev_desc=get_dev(argv[1],dev);
+	if (dev_desc==NULL) {
+		puts ("\n** Invalid device **\n");
+		return 1;
+	}
+
+	if (*ep) {
+		if (*ep != ':') {
+			puts ("\n** Invalid device, use `dev[:part]' **\n");
+			return 1;
+		}
+		part = (int)simple_strtoul(++ep, NULL, 16);
+	}
+
+	if (fat_register_device(dev_desc,part)!=0) {
+		printf ("\n** Unable to use %s %d:%d for fatmv **\n",
+				argv[1],dev,part);
+		return 1;
+	}
+
+	if (file_fat_mv(argv[3], argv[4]) < 0) {
+		printf("\n** Unable to change \"%s\" to \"%s\" **\n",
+			   argv[3], argv[4]);
+		return 1;
+	}
+
+	strncpy(buf, argv[3], sizeof(buf));
+	setenv("filename", buf);
+	return 0;
+}
+
+U_BOOT_CMD(
+	fatmv,	5,	0,	do_fat_fsmv,
+	"fatmv   - move a file in a FAT16/32 filesystem\n",
+	"<interface> <dev[:part]> <oldname> <newname>\n"
+	"    - change <oldname> to <newname> in the FAT16/32 filesytem\n"
+	"      on 'dev' on 'interface'\n"
+);
+
+#endif /* #ifdef CFG_ROCKBOX_FAT */
 
 int do_fat_ls (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
 {
@@ -161,7 +557,8 @@ int do_fat_ls (cmd_tbl_t *cmdtp, int fla
 		part = (int)simple_strtoul(++ep, NULL, 16);
 	}
 	if (fat_register_device(dev_desc,part)!=0) {
-		printf ("\n** Unable to use %s %d:%d for fatls **\n",argv[1],dev,part);
+		printf ("\n** Unable to use %s %d:%d for fatls **\n",
+			argv[1],dev,part);
 		return 1;
 	}
 	if (argc == 4)
@@ -358,5 +755,4 @@ void hexdump (int cnt, unsigned char *da
 	}
 }
 #endif	/* NOT_IMPLEMENTED_YET */
-
 #endif	/* CFG_CMD_FAT */
diff --git a/disk/part.c b/disk/part.c
index 2255e72..b8ecbbf 100644
--- a/disk/part.c
+++ b/disk/part.c
@@ -224,6 +224,8 @@ static void print_part_header (const cha
 					break;
 		case IF_TYPE_DOC:	puts ("DOC");
 					break;
+		case IF_TYPE_SYSACE:	puts ("SystemACE");
+					break;
 		default: 		puts ("UNKNOWN");
 					break;
 	}
diff --git a/disk/part_dos.c b/disk/part_dos.c
index 133ee79..b91cb5e 100644
--- a/disk/part_dos.c
+++ b/disk/part_dos.c
@@ -206,6 +206,9 @@ static int get_partition_info_extended (
 				case IF_TYPE_DOC:
 					sprintf ((char *)info->name, "docd%c%d\n", 'a' + dev_desc->dev, part_num);
 					break;
+				case IF_TYPE_SYSACE:
+					sprintf (info->name, "sysaced%c%d\n", 'a' + dev_desc->dev, part_num);
+					break;
 				default:
 					sprintf ((char *)info->name, "xx%c%d\n", 'a' + dev_desc->dev, part_num);
 					break;
diff --git a/fs/fat/Makefile b/fs/fat/Makefile
index e462757..5d88f2b 100644
--- a/fs/fat/Makefile
+++ b/fs/fat/Makefile
@@ -26,7 +26,8 @@ include $(TOPDIR)/config.mk
 LIB	= libfat.a
 
 AOBJS	=
-COBJS	= fat.o file.o
+COBJS	= fat.o file.o rockbox_fat.o rockbox_dir.o \
+	  rockbox_file.o rockbox_wrapper.o
 
 OBJS	= $(AOBJS) $(COBJS)
 
diff --git a/fs/fat/fat.c b/fs/fat/fat.c
index a823b5a..0f048ed 100644
--- a/fs/fat/fat.c
+++ b/fs/fat/fat.c
@@ -5,6 +5,8 @@
  *
  * 2002-07-28 - rjones@nexus-tech.net - ported to ppcboot v1.1.6
  * 2003-03-10 - kharris@nexus-tech.net - ported to uboot
+ * 2006-01-18 - Keith Outwater (outwater4@comcast.net) - add support for
+ *              rockbox FAT (www.rockbox.org) filesystem driver.
  *
  * See file CREDITS for list of people who contributed to this
  * project.
@@ -33,6 +35,11 @@ #include <part.h>
 
 #if (CONFIG_COMMANDS & CFG_CMD_FAT)
 
+#if defined(CFG_ROCKBOX_FAT)
+extern int rockbox_fat_mount(long startsector);
+#endif
+
+#if !defined(CFG_ROCKBOX_FAT)
 /*
  * Convert a string to lowercase.
  */
@@ -44,26 +51,73 @@ downcase(char *str)
 		str++;
 	}
 }
+#endif /* #if !defined(CFG_ROCKBOX_FAT) */
 
-static  block_dev_desc_t *cur_dev = NULL;
-static unsigned long part_offset = 0;
-static int cur_part = 1;
+cur_block_dev_t cur_block_dev = {
+	cur_dev : NULL,
+	part_offset : 0,
+	cur_part : 1
+};
 
-#define DOS_PART_TBL_OFFSET	0x1be
+#define DOS_PART_TBL_OFFSET		0x1be
 #define DOS_PART_MAGIC_OFFSET	0x1fe
-#define DOS_FS_TYPE_OFFSET	0x36
+#define DOS_FS_TYPE_OFFSET		0x36
 
-int disk_read (__u32 startblock, __u32 getsize, __u8 * bufptr)
+int
+disk_read (__u32 startblock, __u32 getsize, __u8 * bufptr)
 {
-	startblock += part_offset;
-	if (cur_dev == NULL)
+	int rc;
+
+#if !defined(CFG_ROCKBOX_FAT)
+	startblock += cur_block_dev.part_offset;
+#endif
+	if (cur_block_dev.cur_dev == NULL) {
+		FAT_DPRINT("Error: %s:%d:%s: cur_dev is NULL\n",
+				   __FILE__, __LINE__, __FUNCTION__);
 		return -1;
-	if (cur_dev->block_read) {
-		return cur_dev->block_read (cur_dev->dev, startblock, getsize, (unsigned long *)bufptr);
+	}
+	if (cur_block_dev.cur_dev->block_read) {
+		/*
+		 * The block read functions return zero on error, but the FAT 
+		 * filesystem drivers expect a negative return on error.
+		 */
+		rc = (int)cur_block_dev.cur_dev->block_read(cur_block_dev.cur_dev->dev, 
+													startblock, getsize,
+													(unsigned long *)bufptr);
+		if (rc == 0)
+			return -1;
+		else
+			return rc;
 	}
 	return -1;
 }
 
+#if defined(CFG_ROCKBOX_FAT)
+int
+disk_write (__u32 startblock, __u32 putsize, __u8 * bufptr)
+{
+	int rc;
+
+#if !defined(CFG_ROCKBOX_FAT)
+	startblock += cur_block_dev.part_offset;
+#endif
+	if (cur_block_dev.cur_dev == NULL) {
+		FAT_DPRINT("Error: %s:%d:%s: cur_dev is NULL\n",
+				   __FILE__, __LINE__, __FUNCTION__);
+		return -1;
+	}
+	if (cur_block_dev.cur_dev->block_write) {
+		rc = (int)cur_block_dev.cur_dev->block_write(cur_block_dev.cur_dev->dev,
+													 startblock, putsize,
+													 (unsigned long *)bufptr);
+		if (rc == 0)
+			return -1;
+		else
+			return rc;
+	}
+	return -1;
+}
+#endif /* #if defined(CFG_ROCKBOX_FAT) */
 
 int
 fat_register_device(block_dev_desc_t *dev_desc, int part_no)
@@ -72,7 +126,7 @@ fat_register_device(block_dev_desc_t *de
 
 	if (!dev_desc->block_read)
 		return -1;
-	cur_dev=dev_desc;
+	cur_block_dev.cur_dev=dev_desc;
 	/* check if we have a MBR (on floppies we have only a PBR) */
 	if (dev_desc->block_read (dev_desc->dev, 0, 1, (ulong *) buffer) != 1) {
 		printf ("** Can't read from device %d **\n", dev_desc->dev);
@@ -85,19 +139,20 @@ fat_register_device(block_dev_desc_t *de
 	}
 	if(!strncmp((char *)&buffer[DOS_FS_TYPE_OFFSET],"FAT",3)) {
 		/* ok, we assume we are on a PBR only */
-		cur_part = 1;
-		part_offset=0;
+		cur_block_dev.cur_part = 1;
+		cur_block_dev.part_offset=0;
 	}
 	else {
 #if (CONFIG_COMMANDS & CFG_CMD_IDE) || (CONFIG_COMMANDS & CFG_CMD_SCSI) || \
-    (CONFIG_COMMANDS & CFG_CMD_USB) || defined(CONFIG_SYSTEMACE)
+    (CONFIG_COMMANDS & CFG_CMD_USB) || defined(CFG_ROCKBOX_FAT)
 		disk_partition_t info;
 		if(!get_partition_info(dev_desc, part_no, &info)) {
-			part_offset = info.start;
-			cur_part = part_no;
+			cur_block_dev.part_offset = info.start;
+			cur_block_dev.cur_part = part_no;
 		}
 		else {
-			printf ("** Partition %d not valid on device %d **\n",part_no,dev_desc->dev);
+			printf ("** Partition %d not valid on device %d **\n",
+					part_no,dev_desc->dev);
 			return -1;
 		}
 #else
@@ -106,14 +161,21 @@ #else
 		 * by using the get_partition_info routine. For this
 		 * purpose the libpart must be included.
 		 */
-		part_offset=32;
-		cur_part = 1;
+		cur_block_dev.part_offset=32;
+		cur_block_dev.cur_part = 1;
 #endif
 	}
+
+#if defined(CFG_ROCKBOX_FAT)
+	if (rockbox_fat_mount((long)cur_block_dev.part_offset) < 0) {
+		printf("Error: Failed to mount filesystem\n");
+		return -1;
+	}
+#endif
 	return 0;
 }
 
-
+#if !defined(CFG_ROCKBOX_FAT)
 /*
  * Get the first occurence of a directory delimiter ('/' or '\') in a string.
  * Return index into string if found, -1 otherwise.
@@ -129,7 +191,7 @@ dirdelim(char *str)
 	}
 	return -1;
 }
-
+#endif /* #if !defined(CFG_ROCKBOX_FAT) */
 
 /*
  * Match volume_info fs_type strings.
@@ -151,7 +213,7 @@ compare_sign(char *str1, char *str2)
 	return 0;
 }
 
-
+#if !defined(CFG_ROCKBOX_FAT)
 /*
  * Extract zero terminated short name from a directory entry.
  */
@@ -651,7 +713,7 @@ #endif
 
     return NULL;
 }
-
+#endif /* #if !defined(CFG_ROCKBOX_FAT) */
 
 /*
  * Read boot sector and volume info from a FAT filesystem
@@ -712,11 +774,12 @@ #endif
 		}
 	}
 
-	FAT_DPRINT("Error: broken fs_type sign\n");
+	FAT_DPRINT("Error: broken fs_type signature\n");
 	return -1;
 }
 
 
+#if !defined(CFG_ROCKBOX_FAT)
 __u8 do_fat_read_block[MAX_CLUSTSIZE];  /* Block buffer */
 long
 do_fat_read (const char *filename, void *buffer, unsigned long maxsize,
@@ -957,7 +1020,7 @@ #endif
 
     return ret;
 }
-
+#endif /* #if !defined(CFG_ROCKBOX_FAT) */
 
 int
 file_fat_detectfs(void)
@@ -967,24 +1030,26 @@ file_fat_detectfs(void)
 	int		fatsize;
 	char	vol_label[12];
 
-	if(cur_dev==NULL) {
+	if(cur_block_dev.cur_dev==NULL) {
 		printf("No current device\n");
 		return 1;
 	}
 #if (CONFIG_COMMANDS & CFG_CMD_IDE) || (CONFIG_COMMANDS & CFG_CMD_SCSI) || \
-    (CONFIG_COMMANDS & CFG_CMD_USB) || (CONFIG_MMC)
+    (CONFIG_COMMANDS & CFG_CMD_USB) || (CONFIG_MMC) || \
+	((CONFIG_COMMANDS & CFG_CMD_FAT) && defined(CONFIG_SYSTEMACE))
 	printf("Interface:  ");
-	switch(cur_dev->if_type) {
+	switch(cur_block_dev.cur_dev->if_type) {
 		case IF_TYPE_IDE :	printf("IDE"); break;
 		case IF_TYPE_SCSI :	printf("SCSI"); break;
 		case IF_TYPE_ATAPI :	printf("ATAPI"); break;
 		case IF_TYPE_USB :	printf("USB"); break;
 		case IF_TYPE_DOC :	printf("DOC"); break;
 		case IF_TYPE_MMC :	printf("MMC"); break;
+		case IF_TYPE_SYSACE :	printf("SystemACE"); break;
 		default :		printf("Unknown");
 	}
-	printf("\n  Device %d: ",cur_dev->dev);
-	dev_print(cur_dev);
+	printf("\n  Device %d: ",cur_block_dev.cur_dev->dev);
+	dev_print(cur_block_dev.cur_dev);
 #endif
 	if(read_bootsectandvi(&bs, &volinfo, &fatsize)) {
 		printf("\nNo valid FAT fs found\n");
@@ -993,11 +1058,11 @@ #endif
 	memcpy (vol_label, volinfo.volume_label, 11);
 	vol_label[11] = '\0';
 	volinfo.fs_type[5]='\0';
-	printf("Partition %d: Filesystem: %s \"%s\"\n",cur_part,volinfo.fs_type,vol_label);
+	printf("Partition %d: Filesystem: %s \"%s\"\n",cur_block_dev.cur_part,volinfo.fs_type,vol_label);
 	return 0;
 }
 
-
+#if !defined(CFG_ROCKBOX_FAT)
 int
 file_fat_ls(const char *dir)
 {
@@ -1011,5 +1076,5 @@ file_fat_read(const char *filename, void
 	printf("reading %s\n",filename);
 	return do_fat_read(filename, buffer, maxsize, LS_NO);
 }
-
+#endif /* #if !defiined(CFG_ROCKBOX_FAT) */
 #endif /* #if (CONFIG_COMMANDS & CFG_CMD_FAT) */
diff --git a/fs/fat/file.c b/fs/fat/file.c
index f999ac5..386b282 100644
--- a/fs/fat/file.c
+++ b/fs/fat/file.c
@@ -5,6 +5,7 @@
  *
  * 2002-07-28 - rjones@nexus-tech.net - ported to ppcboot v1.1.6
  * 2003-03-10 - kharris@nexus-tech.net - ported to uboot
+ * 2006-01-18 - Keith Outwater (outwater@comcast.net) - added write support
  *
  * See file CREDITS for list of people who contributed to this
  * project.
@@ -34,10 +35,22 @@ #include <linux/time.h>
 
 #if (CONFIG_COMMANDS & CFG_CMD_FAT)
 
+#include "rockbox_debug.h"
+
+#ifdef CFG_ROCKBOX_FAT
+extern cur_block_dev_t cur_block_dev;
+#endif
+
+#if !defined(CFG_ROCKBOX_FAT)
 /* Supported filesystems */
 static const struct filesystem filesystems[] = {
-	{ file_fat_detectfs,  file_fat_ls,  file_fat_read,  "FAT" },
+	{ detect:	file_fat_detectfs,
+	  ls:		file_fat_ls,
+	  read:		file_fat_read,
+	  "FAT"
+	},
 };
+
 #define NUM_FILESYS	(sizeof(filesystems)/sizeof(struct filesystem))
 
 /* The filesystem which was last detected */
@@ -140,10 +153,10 @@ file_cd(const char *path)
 		pathcpy(tmpstr+1, path);
 	}
 
+	fat_dprintf("New CWD is '%s'\n", file_cwd);
 	return 0;
 }
 
-
 int
 file_detectfs(void)
 {
@@ -205,4 +218,5 @@ file_read(const char *filename, void *bu
 	return filesystems[current_filesystem].read(arg, buffer, maxsize);
 }
 
+#endif /* #if !defined(CFG_ROCKBOX_FAT) */
 #endif /* #if (CONFIG_COMMANDS & CFG_CMD_FAT) */
diff --git a/fs/fat/rockbox_debug.h b/fs/fat/rockbox_debug.h
new file mode 100644
index 0000000..10d4a6e
--- /dev/null
+++ b/fs/fat/rockbox_debug.h
@@ -0,0 +1,39 @@
+/*------------------------------------------------------------------------------
+ * (C) Copyright 2006
+ * Keith Outwater, (outwater4@comcast.net)
+ *
+ * See file CREDITS for list of people who contributed to this
+ * project.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without eve8 the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+ * MA 02111-1307 USA
+ */
+
+#define	fat_eprintf(fmt,args...) fprintf(stderr, "Error: %s:%d:%s(): "fmt"", \
+					__FILE__, __LINE__,__FUNCTION__,##args)
+#define	fat_pprintf(fmt,args...) fprintf(stderr, "Fatal: %s:%d:%s(): "fmt"", \
+					__FILE__, __LINE__,__FUNCTION__,##args)
+#define	fat_wprintf(fmt,args...) fprintf(stderr, "Warning: %s:%d:%s(): "fmt"", \
+					__FILE__, __LINE__,__FUNCTION__,##args)
+
+#if (defined(DEBUG) || defined(FAT_DEBUG))
+#define	fat_dprintf(fmt,args...) printf("Debug: %s:%d:%s(): "fmt"", \
+					__FILE__, __LINE__,__FUNCTION__,##args)
+#define	fat_deprintf(fmt,args...) fprintf(stderr, "Error: %s:%d:%s(): "fmt"",\
+					__FILE__, __LINE__,__FUNCTION__,##args)
+#else
+#define	fat_dprintf(fmt,args...)
+#define	fat_deprintf(fmt,args...)
+#endif
diff --git a/fs/fat/rockbox_dir.c b/fs/fat/rockbox_dir.c
new file mode 100644
index 0000000..2e6f400
--- /dev/null
+++ b/fs/fat/rockbox_dir.c
@@ -0,0 +1,250 @@
+/***************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: dir.c,v 1.28 2005/01/28 21:32:16 hohensoh Exp $
+ *
+ * Copyright (C) 2002 by Bjrn Stenberg
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.29 of 'firmware/common/dir.c' from rockbox CVS server.
+ *
+ ****************************************************************************/
+
+#include <common.h>
+#include <config.h>
+
+#if defined(CFG_ROCKBOX_FAT)
+
+#include <asm/errno.h>
+#include <linux/string.h>
+#include <fat.h>
+#include "rockbox_debug.h"
+
+#ifndef __HAVE_ARCH_STRNICMP
+extern int strnicmp(const char *s1, const char *s2, size_t len);
+#endif
+
+#if !defined(CFG_ROCKBOX_FAT_MAX_OPEN_DIRS)
+#define CFG_ROCKBOX_FAT_MAX_OPEN_DIRS 8
+#endif
+
+static DIR opendirs[CFG_ROCKBOX_FAT_MAX_OPEN_DIRS];
+
+DIR *
+opendir(const char *name)
+{
+	char namecopy[MAX_PATH];
+	char *part;
+	struct fat_direntry entry;
+	int dd;
+	DIR *pdir = opendirs;
+
+	if (name[0] != '/') {
+		fprintf(stderr, "Error: Only absolute paths supported\n");
+		return NULL;
+	}
+
+	/*
+	 * find a free dir descriptor 
+	 */
+	for (dd = 0; dd < CFG_ROCKBOX_FAT_MAX_OPEN_DIRS; dd++, pdir++)
+		if (!pdir->busy)
+			break;
+
+	if (dd == CFG_ROCKBOX_FAT_MAX_OPEN_DIRS) {
+		fat_eprintf("Too many dirs open\n");
+		errno = EMFILE;
+		return NULL;
+	}
+
+	pdir->busy = true;
+
+	strncpy(namecopy, name, sizeof (namecopy));	/* just copy */
+	namecopy[sizeof (namecopy) - 1] = '\0';
+
+	if (fat_opendir(&pdir->fatdir, 0, NULL) < 0) {
+		fat_deprintf("Failed to open root dir\n");
+		pdir->busy = false;
+		return NULL;
+	}
+
+	for (part = strtok(namecopy, "/"); part; part = strtok(NULL, "/")) {
+		/*
+		 * scan dir for name 
+		 */
+		while (1) {
+			if ((fat_getnext(&pdir->fatdir, &entry) < 0) ||
+			    (!entry.name[0])) {
+				pdir->busy = false;
+				return NULL;
+			}
+			if ((entry.attr & FAT_ATTR_DIRECTORY) &&
+			    (!strnicmp(part, entry.name, strlen(part)))) {
+				pdir->parent_dir = pdir->fatdir;
+				if (fat_opendir(&pdir->fatdir,
+						entry.firstcluster,
+						&pdir->parent_dir) < 0) {
+					fat_deprintf
+					    ("Failed to open dir \"%s\" "
+					     "(firstcluster = %ld)\n", part,
+					     entry.firstcluster);
+					pdir->busy = false;
+					return NULL;
+				}
+				break;
+			}
+		}
+	}
+
+	return pdir;
+}
+
+int
+closedir(DIR * dir)
+{
+	dir->busy = false;
+	return 0;
+}
+
+struct dirent *
+readdir(DIR * dir)
+{
+	struct fat_direntry entry;
+	struct dirent *theent = &(dir->theent);
+
+	if (!dir->busy)
+		return NULL;
+
+	/*
+	 * normal directory entry fetching follows here 
+	 */
+	if (fat_getnext(&(dir->fatdir), &entry) < 0)
+		return NULL;
+
+	if (!entry.name[0])
+		return NULL;
+
+	strncpy(theent->d_name, entry.name, sizeof (theent->d_name));
+	theent->attribute = entry.attr;
+	theent->size = entry.filesize;
+	theent->startcluster = entry.firstcluster;
+	theent->wrtdate = entry.wrtdate;
+	theent->wrttime = entry.wrttime;
+
+	return theent;
+}
+
+int
+mkdir(const char *name, int mode)
+{
+	DIR *dir;
+	char namecopy[MAX_PATH];
+	char *end;
+	char *basename;
+	char *parent;
+	struct dirent *entry;
+	struct fat_dir newdir;
+	int rc;
+
+	(void) mode;
+
+	if (name[0] != '/') {
+		fprintf(stderr, "Error: Only absolute paths supported\n");
+		return -1;
+	}
+
+	strncpy(namecopy, name, sizeof (namecopy));
+	namecopy[sizeof (namecopy) - 1] = 0;
+
+	/*
+	 * Split the base name and the path 
+	 */
+	end = strrchr(namecopy, '/');
+	*end = 0;
+	basename = end + 1;
+
+	if (namecopy == end)	/* Root dir? */
+		parent = "/";
+	else
+		parent = namecopy;
+
+	fat_dprintf("parent: \"%s\", name: \"%s\"\n", parent, basename);
+
+	dir = opendir(parent);
+
+	if (!dir) {
+		fprintf(stderr, "Can't open parent dir\n");
+		return -2;
+	}
+
+	if (basename[0] == 0) {
+		fat_eprintf("Empty directory name\n");
+		errno = EINVAL;
+		return -3;
+	}
+
+	/*
+	 * Now check if the name already exists 
+	 */
+	while ((entry = readdir(dir))) {
+		if (!strnicmp(basename, entry->d_name, strlen(basename))) {
+			fat_eprintf("File exists\n");
+			errno = EEXIST;
+			closedir(dir);
+			return -4;
+		}
+	}
+
+	memset(&newdir, sizeof (struct fat_dir), 0);
+	rc = fat_create_dir(basename, &newdir, &(dir->fatdir));
+	closedir(dir);
+	return rc;
+}
+
+int
+rmdir(const char *name)
+{
+	int rc;
+	DIR *dir;
+	struct dirent *entry;
+
+	dir = opendir(name);
+	if (!dir) {
+		errno = ENOENT;	/* open error */
+		return -1;
+	}
+
+	/*
+	 * check if the directory is empty 
+	 */
+	while ((entry = readdir(dir))) {
+		if (strcmp(entry->d_name, ".") && strcmp(entry->d_name, "..")) {
+			fat_eprintf("Directory is not empty\n");
+			errno = ENOTEMPTY;
+			closedir(dir);
+			return -2;
+		}
+	}
+
+	rc = fat_remove(&(dir->fatdir.file));
+	if (rc < 0) {
+		fat_eprintf("Failed to remove directory (rc = %d)\n", rc);
+		errno = EIO;
+		rc = rc * 10 - 3;
+	}
+
+	closedir(dir);
+	return rc;
+}
+
+#endif				/* #if defined(CFG_ROCKBOX_FAT) */
diff --git a/fs/fat/rockbox_fat.c b/fs/fat/rockbox_fat.c
new file mode 100644
index 0000000..cdc1638
--- /dev/null
+++ b/fs/fat/rockbox_fat.c
@@ -0,0 +1,2489 @@
+/*******************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: fat.c,v 1.119 2005/09/07 01:35:15 rasher Exp $
+ *
+ * Copyright (C) 2002 by Linus Nielsen Feltzing
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.123 of 'firmware/drivers/fat.c' from rockbox CVS server.
+ *
+ ****************************************************************************/
+
+#include <common.h>
+#include <config.h>
+
+#if defined(CFG_ROCKBOX_FAT)
+#include <linux/ctype.h>
+#include <fat.h>
+#include "rockbox_debug.h"
+
+/*
+ * Override this parameter to boost performance ONLY if you are sure your driver
+ * can handle requests of more than 256 sectors!
+ */
+#if !defined(CFG_ROCKBOX_FAT_MAX_SECS_PER_XFER)
+#define CFG_ROCKBOX_FAT_MAX_SECS_PER_XFER	256
+#endif
+
+#define	SWAB16	FAT2CPU16
+#define	SWAB32	FAT2CPU32
+
+#if !defined(NO_ROCKBOX_FAT16_SUPPORT)
+#define	HAVE_FAT16SUPPORT
+#else
+#undef	HAVE_FAT16SUPPORT
+#endif
+
+extern int disk_read(__u32 startblock, __u32 getsize, __u8 * bufptr);
+extern int disk_write(__u32 startblock, __u32 putsize, __u8 * bufptr);
+
+/*
+ * The current working directory.
+ */
+extern char file_cwd[];
+
+/*
+ * The current device and partition information. Updated by
+ * fat_register_device().
+ */
+extern cur_block_dev_t cur_block_dev;
+
+#define BYTES2INT16(array,pos) \
+          (array[pos] | (array[pos+1] << 8 ))
+#define BYTES2INT32(array,pos)					\
+    ((long)array[pos] | ((long)array[pos+1] << 8 ) |		\
+     ((long)array[pos+2] << 16 ) | ((long)array[pos+3] << 24 ))
+
+#define FATTYPE_FAT12       0
+#define FATTYPE_FAT16       1
+#define FATTYPE_FAT32       2
+
+/* BPB offsets; generic */
+#define BS_JMPBOOT          0
+#define BS_OEMNAME          3
+#define BPB_BYTSPERSEC      11
+#define BPB_SECPERCLUS      13
+#define BPB_RSVDSECCNT      14
+#define BPB_NUMFATS         16
+#define BPB_ROOTENTCNT      17
+#define BPB_TOTSEC16        19
+#define BPB_MEDIA           21
+#define BPB_FATSZ16         22
+#define BPB_SECPERTRK       24
+#define BPB_NUMHEADS        26
+#define BPB_HIDDSEC         28
+#define BPB_TOTSEC32        32
+
+/* fat12/16 */
+#define BS_DRVNUM           36
+#define BS_RESERVED1        37
+#define BS_BOOTSIG          38
+#define BS_VOLID            39
+#define BS_VOLLAB           43
+#define BS_FILSYSTYPE       54
+
+/* fat32 */
+#define BPB_FATSZ32         36
+#define BPB_EXTFLAGS        40
+#define BPB_FSVER           42
+#define BPB_ROOTCLUS        44
+#define BPB_FSINFO          48
+#define BPB_BKBOOTSEC       50
+#define BS_32_DRVNUM        64
+#define BS_32_BOOTSIG       66
+#define BS_32_VOLID         67
+#define BS_32_VOLLAB        71
+#define BS_32_FILSYSTYPE    82
+
+#define BPB_LAST_WORD       510
+
+/* attributes */
+#define FAT_ATTR_LONG_NAME   (FAT_ATTR_READ_ONLY | FAT_ATTR_HIDDEN | \
+                              FAT_ATTR_SYSTEM | FAT_ATTR_VOLUME_ID)
+#define FAT_ATTR_LONG_NAME_MASK (FAT_ATTR_READ_ONLY | FAT_ATTR_HIDDEN | \
+                                 FAT_ATTR_SYSTEM | FAT_ATTR_VOLUME_ID | \
+                                 FAT_ATTR_DIRECTORY | FAT_ATTR_ARCHIVE )
+
+/* NTRES flags */
+#define FAT_NTRES_LC_NAME    0x08
+#define FAT_NTRES_LC_EXT     0x10
+
+#define FATDIR_NAME          0
+#define FATDIR_ATTR          11
+#define FATDIR_NTRES         12
+#define FATDIR_CRTTIMETENTH  13
+#define FATDIR_CRTTIME       14
+#define FATDIR_CRTDATE       16
+#define FATDIR_LSTACCDATE    18
+#define FATDIR_FSTCLUSHI     20
+#define FATDIR_WRTTIME       22
+#define FATDIR_WRTDATE       24
+#define FATDIR_FSTCLUSLO     26
+#define FATDIR_FILESIZE      28
+
+#define FATLONG_ORDER        0
+#define FATLONG_TYPE         12
+#define FATLONG_CHKSUM       13
+
+#define CLUSTERS_PER_FAT_SECTOR (SECTOR_SIZE / 4)
+#define CLUSTERS_PER_FAT16_SECTOR (SECTOR_SIZE / 2)
+#define DIR_ENTRIES_PER_SECTOR  (SECTOR_SIZE / DIR_ENTRY_SIZE)
+#define DIR_ENTRY_SIZE       32
+#define NAME_BYTES_PER_ENTRY 13
+#define FAT_BAD_MARK         0x0ffffff7
+#define FAT_EOF_MARK         0x0ffffff8
+
+/* filename charset conversion table */
+static const unsigned char unicode2iso8859_2[] = {
+	0x00, 0x00, 0xc3, 0xe3, 0xa1, 0xb1, 0xc6, 0xe6,	/* 0x0100 */
+	0x00, 0x00, 0x00, 0x00, 0xc8, 0xe8, 0xcf, 0xef,	/* 0x0108 */
+	0xd0, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0110 */
+	0xca, 0xea, 0xcc, 0xec, 0x00, 0x00, 0x00, 0x00,	/* 0x0118 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0120 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0128 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0130 */
+	0x00, 0xc5, 0xe5, 0x00, 0x00, 0xa5, 0xb5, 0x00,	/* 0x0138 */
+	0x00, 0xa3, 0xb3, 0xd1, 0xf1, 0x00, 0x00, 0xd2,	/* 0x0140 */
+	0xf2, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0148 */
+	0xd5, 0xf5, 0x00, 0x00, 0xc0, 0xe0, 0x00, 0x00,	/* 0x0150 */
+	0xd8, 0xf8, 0xa6, 0xb6, 0x00, 0x00, 0xaa, 0xba,	/* 0x0158 */
+	0xa9, 0xb9, 0xde, 0xfe, 0xab, 0xbb, 0x00, 0x00,	/* 0x0160 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xd9, 0xf9,	/* 0x0168 */
+	0xdb, 0xfb, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0170 */
+	0x00, 0xac, 0xbc, 0xaf, 0xbf, 0xae, 0xbe, 0x00,	/* 0x0178 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0180 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0188 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0190 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0198 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01a0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01a8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01b0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01b8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01c0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01c8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01d0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01d8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01e0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01e8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01f0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x01f8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0200 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0208 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0210 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0218 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0220 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0228 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0230 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0238 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0240 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0248 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0250 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0258 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0260 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0268 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0270 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0278 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0280 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0288 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0290 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x0298 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02a0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02a8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02b0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02b8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xb7,	/* 0x02c0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02c8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02d0 */
+	0xa2, 0xff, 0x00, 0xb2, 0x00, 0xbd, 0x00, 0x00,	/* 0x02d8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02e0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02e8 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x02f0 */
+	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00	/* 0x02f8 */
+};
+
+struct fsinfo {
+	unsigned long freecount;	/* last known free cluster count */
+	unsigned long nextfree;	/* first cluster to start looking for free
+				 * clusters, or 0xffffffff for no hint */
+};
+
+/* fsinfo offsets */
+#define FSINFO_FREECOUNT 488
+#define FSINFO_NEXTFREE  492
+
+struct bpb {
+	int bpb_bytspersec;	/* Bytes per sector, typically 512 */
+	unsigned int bpb_secperclus;	/* Sectors per cluster */
+	int bpb_rsvdseccnt;	/* Number of reserved sectors */
+	int bpb_numfats;	/* Number of FAT structures, typically 2 */
+	int bpb_totsec16;	/* Number of sectors on volume (old 16-bit) */
+	int bpb_media;		/* Media type (typically 0xf0 or 0xf8) */
+	int bpb_fatsz16;	/* Number of used sectors per FAT structure */
+	unsigned long bpb_totsec32;	/* Num sectson the volume (new 32-bit) */
+	unsigned int last_word;	/* 0xAA55 */
+
+    /**** FAT32 specific *****/
+	long bpb_fatsz32;
+	long bpb_rootclus;
+	long bpb_fsinfo;
+
+	/*
+	 * variables for internal use 
+	 */
+	unsigned long fatsize;
+	unsigned long totalsectors;
+	unsigned long rootdirsector;
+	unsigned long firstdatasector;
+	unsigned long startsector;
+	unsigned long dataclusters;
+	struct fsinfo fsinfo;
+#ifdef HAVE_FAT16SUPPORT
+	int bpb_rootentcnt;	/* Number of dir entries in the root */
+	/*
+	 * internals for FAT16 support 
+	 */
+	bool is_fat16;		/* true if we mounted a FAT16 partition, false
+				 * if FAT32 */
+	/*
+	 * sect offset of root dir relative to start of first pseudo cluster 
+	 */
+	unsigned int rootdiroffset;
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+};
+
+/*
+ * mounted partition info
+ */
+static struct bpb fat_bpb;
+
+static int update_fsinfo(void);
+static int flush_fat(void);
+static int bpb_is_sane(void);
+static void *cache_fat_sector(long secnum, bool dirty);
+static void create_dos_name(const unsigned char *name, unsigned char *newname);
+static void randomize_dos_name(unsigned char *name);
+static unsigned long find_free_cluster(unsigned long start);
+static int transfer(unsigned long start, long count, char *buf, bool write);
+
+#define FAT_CACHE_SIZE 0x20
+#define FAT_CACHE_MASK (FAT_CACHE_SIZE-1)
+
+struct fat_cache_entry {
+	long secnum;
+	bool inuse;
+	bool dirty;
+};
+
+static char fat_cache_sectors[FAT_CACHE_SIZE][SECTOR_SIZE];
+static struct fat_cache_entry fat_cache[FAT_CACHE_SIZE];
+
+static long
+cluster2sec(long cluster)
+{
+#ifdef HAVE_FAT16SUPPORT
+	/*
+	 * negative clusters (FAT16 root dir) don't get the 2 offset 
+	 */
+	int zerocluster = cluster < 0 ? 0 : 2;
+#else
+	const long zerocluster = 2;
+#endif
+
+	if (cluster > (long) (fat_bpb.dataclusters + 1)) {
+		fat_eprintf("Bad cluster number (cluster %ld)\n", cluster);
+		return -1;
+	}
+
+	return (cluster - zerocluster) * fat_bpb.bpb_secperclus
+	    + fat_bpb.firstdatasector;
+}
+
+void
+fat_size(unsigned long *size, unsigned long *free)
+{
+	if (size)
+		*size = fat_bpb.dataclusters * fat_bpb.bpb_secperclus / 2;
+
+	if (free)
+		*free = fat_bpb.fsinfo.freecount * fat_bpb.bpb_secperclus / 2;
+}
+
+static void
+fat_init(void)
+{
+	unsigned int i;
+
+	/*
+	 * mark the FAT cache as unused 
+	 */
+	for (i = 0; i < FAT_CACHE_SIZE; i++) {
+		fat_cache[i].secnum = 8;	/* We use a "safe" sector just
+						 * in case */
+		fat_cache[i].inuse = false;
+		fat_cache[i].dirty = false;
+	}
+}
+
+/*
+ * Mount the FAT filesystem located at 'startsector'.
+ * Return 0 on success, an error-specific negative number on error.
+ */
+int
+rockbox_fat_mount(long startsector)
+{
+	unsigned char buf[SECTOR_SIZE];
+	int rc;
+	long datasec;
+#ifdef HAVE_FAT16SUPPORT
+	int rootdirsectors;
+#endif
+
+	fat_dprintf("startsector %ld\n", startsector);
+	fat_init();
+	rc = disk_read(startsector, 1, buf);
+
+	if (rc < 0) {
+		fat_eprintf("Could not read BPB (rc = %d)\n", rc);
+		return rc * 10 - 1;
+	}
+
+	memset(&fat_bpb, 0, sizeof (struct bpb));
+	fat_bpb.startsector = startsector;
+
+	fat_bpb.bpb_bytspersec = BYTES2INT16(buf, BPB_BYTSPERSEC);
+	fat_bpb.bpb_secperclus = buf[BPB_SECPERCLUS];
+	fat_bpb.bpb_rsvdseccnt = BYTES2INT16(buf, BPB_RSVDSECCNT);
+	fat_bpb.bpb_numfats = buf[BPB_NUMFATS];
+	fat_bpb.bpb_totsec16 = BYTES2INT16(buf, BPB_TOTSEC16);
+	fat_bpb.bpb_media = buf[BPB_MEDIA];
+	fat_bpb.bpb_fatsz16 = BYTES2INT16(buf, BPB_FATSZ16);
+	fat_bpb.bpb_fatsz32 = BYTES2INT32(buf, BPB_FATSZ32);
+	fat_bpb.bpb_totsec32 = BYTES2INT32(buf, BPB_TOTSEC32);
+	fat_bpb.last_word = BYTES2INT16(buf, BPB_LAST_WORD);
+
+	/*
+	 * calculate a few commonly used values 
+	 */
+	if (fat_bpb.bpb_fatsz16 != 0)
+		fat_bpb.fatsize = fat_bpb.bpb_fatsz16;
+	else
+		fat_bpb.fatsize = fat_bpb.bpb_fatsz32;
+
+	if (fat_bpb.bpb_totsec16 != 0)
+		fat_bpb.totalsectors = fat_bpb.bpb_totsec16;
+	else
+		fat_bpb.totalsectors = fat_bpb.bpb_totsec32;
+
+#ifdef HAVE_FAT16SUPPORT
+	fat_bpb.bpb_rootentcnt = BYTES2INT16(buf, BPB_ROOTENTCNT);
+	rootdirsectors = ((fat_bpb.bpb_rootentcnt * 32)
+			  + (fat_bpb.bpb_bytspersec -
+			     1)) / fat_bpb.bpb_bytspersec;
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+
+	fat_bpb.firstdatasector = fat_bpb.bpb_rsvdseccnt
+#ifdef HAVE_FAT16SUPPORT
+	    + rootdirsectors
+#endif
+	    + fat_bpb.bpb_numfats * fat_bpb.fatsize;
+
+	/*
+	 * Determine FAT type 
+	 */
+	datasec = fat_bpb.totalsectors - fat_bpb.firstdatasector;
+	fat_bpb.dataclusters = datasec / fat_bpb.bpb_secperclus;
+
+	if (fat_bpb.dataclusters < 65525) {	/* FAT16 */
+#ifdef HAVE_FAT16SUPPORT
+		fat_bpb.is_fat16 = true;
+		if (fat_bpb.dataclusters < 4085) {	/* FAT12 */
+			fat_eprintf("FAT12 filesystem not supported\n");
+			return -2;
+		}
+#else				/* #ifdef HAVE_FAT16SUPPORT */
+		fat_eprintf("Not a FAT32 filesystem\n");
+		return -2;
+#endif				/* #ifndef HAVE_FAT16SUPPORT */
+	}
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {	/* FAT16 specific part of BPB */
+		int dirclusters;
+		fat_bpb.rootdirsector = fat_bpb.bpb_rsvdseccnt
+		    + fat_bpb.bpb_numfats * fat_bpb.bpb_fatsz16;
+		dirclusters = ((rootdirsectors + fat_bpb.bpb_secperclus - 1)
+			       / fat_bpb.bpb_secperclus);	/* rounded up,
+								 * to full
+								 * clusters */
+		/*
+		 * I assign negative pseudo cluster numbers for the root
+		 * directory, their range is counted upward until -1. 
+		 */
+		fat_bpb.bpb_rootclus = 0 - dirclusters;	/* backwards, before
+							 * the data */
+		fat_bpb.rootdiroffset = dirclusters * fat_bpb.bpb_secperclus
+		    - rootdirsectors;
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{			/* FAT32 specific part of BPB */
+		fat_bpb.bpb_rootclus = BYTES2INT32(buf, BPB_ROOTCLUS);
+		fat_bpb.bpb_fsinfo = BYTES2INT16(buf, BPB_FSINFO);
+		fat_bpb.rootdirsector = cluster2sec(fat_bpb.bpb_rootclus);
+	}
+
+	rc = bpb_is_sane();
+	if (rc < 0) {
+		fat_eprintf("BPB is not sane\n");
+		return rc * 10 - 3;
+	}
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		fat_bpb.fsinfo.freecount = 0xffffffff;	/* force recalc below */
+		fat_bpb.fsinfo.nextfree = 0xffffffff;
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{
+		/*
+		 * Read the fsinfo sector 
+		 */
+		rc = disk_read(startsector + fat_bpb.bpb_fsinfo, 1, buf);
+		if (rc < 0) {
+			fat_eprintf("Could not read FSInfo (rc = %d)\n", rc);
+			return rc * 10 - 4;
+		}
+		fat_bpb.fsinfo.freecount = BYTES2INT32(buf, FSINFO_FREECOUNT);
+		fat_bpb.fsinfo.nextfree = BYTES2INT32(buf, FSINFO_NEXTFREE);
+	}
+
+	/*
+	 * calculate freecount if unset 
+	 */
+	if (fat_bpb.fsinfo.freecount == 0xffffffff) {
+		fat_recalc_free();
+	}
+
+	fat_dprintf("\n  freecount %ld, nextfree 0x%lx, cluster count 0x%lx\n"
+		    "  sectors/cluster %d, FAT sectors 0x%lx\n",
+		    fat_bpb.fsinfo.freecount, fat_bpb.fsinfo.nextfree,
+		    fat_bpb.dataclusters, fat_bpb.bpb_secperclus,
+		    fat_bpb.fatsize);
+	return 0;
+}
+
+void
+fat_recalc_free(void)
+{
+	long free = 0;
+	unsigned long i;
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		for (i = 0; i < fat_bpb.fatsize; i++) {
+			unsigned int j;
+			unsigned short *fat = cache_fat_sector(i, false);
+			for (j = 0; j < CLUSTERS_PER_FAT16_SECTOR; j++) {
+				unsigned int c =
+				    i * CLUSTERS_PER_FAT16_SECTOR + j;
+				if (c > fat_bpb.dataclusters + 1)	/* nr 0 
+									 * is
+									 * unused 
+									 */
+					break;
+
+				if (SWAB16(fat[j]) == 0x0000) {
+					free++;
+					if (fat_bpb.fsinfo.nextfree ==
+					    0xffffffff)
+						fat_bpb.fsinfo.nextfree = c;
+				}
+			}
+		}
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{
+		for (i = 0; i < fat_bpb.fatsize; i++) {
+			unsigned int j;
+			unsigned long *fat = cache_fat_sector(i, false);
+			for (j = 0; j < CLUSTERS_PER_FAT_SECTOR; j++) {
+				unsigned long c =
+				    i * CLUSTERS_PER_FAT_SECTOR + j;
+				if (c > fat_bpb.dataclusters + 1)	/* nr 0 
+									 * is
+									 * unused 
+									 */
+					break;
+
+				if (!(SWAB32(fat[j]) & 0x0fffffff)) {
+					free++;
+					if (fat_bpb.fsinfo.nextfree ==
+					    0xffffffff)
+						fat_bpb.fsinfo.nextfree = c;
+				}
+			}
+		}
+	}
+	fat_bpb.fsinfo.freecount = free;
+	update_fsinfo();
+}
+
+static int
+bpb_is_sane(void)
+{
+
+	if (fat_bpb.bpb_bytspersec != SECTOR_SIZE) {
+		fat_eprintf("Sector size is not %d (detected %d)\n",
+			    SECTOR_SIZE, fat_bpb.bpb_bytspersec);
+		return -1;
+	}
+
+	if ((long) fat_bpb.bpb_secperclus * (long) fat_bpb.bpb_bytspersec >
+	    128L * 1024L) {
+		fat_eprintf("Cluster size larger than 128K (computed %d)\n",
+			    fat_bpb.bpb_bytspersec * fat_bpb.bpb_secperclus);
+		return -2;
+	}
+
+	if (fat_bpb.bpb_numfats != 2) {
+		printf("Warning: Number of FATs is not 2 (detected %d)\n",
+		       fat_bpb.bpb_numfats);
+	}
+
+	if (fat_bpb.bpb_media != 0xf0 && fat_bpb.bpb_media < 0xf8) {
+		printf("Warning: Non-standard media type (detected 0x%02x)\n",
+		       fat_bpb.bpb_media);
+	}
+
+	if (fat_bpb.last_word != 0xaa55) {
+		fat_eprintf("Last word is not 0xaa55 (detected 0x%04x)\n",
+			    fat_bpb.last_word);
+		return -3;
+	}
+
+	if (fat_bpb.fsinfo.freecount >
+	    (fat_bpb.totalsectors - fat_bpb.firstdatasector) /
+	    fat_bpb.bpb_secperclus) {
+		fat_eprintf("FSInfo.Freecount > disk size "
+			    "(freecount = 0x%04lx)\n",
+			    fat_bpb.fsinfo.freecount);
+		return -4;
+	}
+
+	return 0;
+}
+
+static void
+flush_fat_sector(struct fat_cache_entry *fce, unsigned char *sectorbuf)
+{
+	int rc;
+	long secnum;
+
+	secnum = fce->secnum + fat_bpb.startsector;
+
+	/*
+	 * Write to the first FAT 
+	 */
+	rc = disk_write(secnum, 1, sectorbuf);
+	if (rc < 0) {
+		fat_pprintf("Could not write sector %ld (rc = %d)\n", secnum,
+			    rc);
+	}
+
+	if (fat_bpb.bpb_numfats > 1) {
+		/*
+		 * Write to the second FAT 
+		 */
+		secnum += fat_bpb.fatsize;
+		rc = disk_write(secnum, 1, sectorbuf);
+		if (rc < 0) {
+			fat_pprintf("Could not write sector %ld (rc = %d)\n",
+				    secnum, rc);
+		}
+	}
+
+	fce->dirty = false;
+}
+
+static void *
+cache_fat_sector(long fatsector, bool dirty)
+{
+	long secnum = fatsector + fat_bpb.bpb_rsvdseccnt;
+	int cache_index = secnum & FAT_CACHE_MASK;
+	struct fat_cache_entry *fce = &fat_cache[cache_index];
+	unsigned char *sectorbuf = &fat_cache_sectors[cache_index][0];
+	int rc;
+
+	/*
+	 * Delete the cache entry if it isn't the sector we want 
+	 */
+	if (fce->inuse && (fce->secnum != secnum)) {
+		/*
+		 * Write back if it is dirty 
+		 */
+		if (fce->dirty) {
+			flush_fat_sector(fce, sectorbuf);
+		}
+		fce->inuse = false;
+	}
+
+	/*
+	 * Load the sector if it is not cached 
+	 */
+	if (!fce->inuse) {
+		rc = disk_read(secnum + fat_bpb.startsector, 1, sectorbuf);
+		if (rc < 0) {
+			fprintf(stderr, "Read of sector %ld failed (rc = %d)\n",
+				secnum, rc);
+			return NULL;
+		}
+
+		fce->inuse = true;
+		fce->secnum = secnum;
+	}
+
+	if (dirty)
+		fce->dirty = true;	/* dirt remains, sticky until flushed */
+
+	return sectorbuf;
+}
+
+static unsigned long
+find_free_cluster(unsigned long startcluster)
+{
+	unsigned long sector;
+	unsigned long offset;
+	unsigned long i;
+
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		sector = startcluster / CLUSTERS_PER_FAT16_SECTOR;
+		offset = startcluster % CLUSTERS_PER_FAT16_SECTOR;
+
+		for (i = 0; i < fat_bpb.fatsize; i++) {
+			unsigned int j;
+			unsigned int nr = (i + sector) % fat_bpb.fatsize;
+			unsigned short *fat = cache_fat_sector(nr, false);
+			if (!fat)
+				break;
+
+			for (j = 0; j < CLUSTERS_PER_FAT16_SECTOR; j++) {
+				int k =
+				    (j + offset) % CLUSTERS_PER_FAT16_SECTOR;
+				if (SWAB16(fat[k]) == 0x0000) {
+					unsigned int c =
+					    nr * CLUSTERS_PER_FAT16_SECTOR + k;
+					/*
+					 * Ignore the reserved clusters 0 & 1,
+					 * and also cluster numbers out of
+					 * bounds 
+					 */
+					if (c < 2
+					    || c > fat_bpb.dataclusters + 1) {
+						continue;
+					}
+
+					fat_dprintf("startcluster %lu, c %u\n",
+						    startcluster, c);
+					fat_bpb.fsinfo.nextfree = c;
+					return c;
+				}
+			}
+			offset = 0;
+		}
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{
+		sector = startcluster / CLUSTERS_PER_FAT_SECTOR;
+		offset = startcluster % CLUSTERS_PER_FAT_SECTOR;
+
+		for (i = 0; i < fat_bpb.fatsize; i++) {
+			unsigned int j;
+			unsigned long nr = (i + sector) % fat_bpb.fatsize;
+			unsigned long *fat = cache_fat_sector(nr, false);
+			if (!fat)
+				break;
+
+			for (j = 0; j < CLUSTERS_PER_FAT_SECTOR; j++) {
+				int k = (j + offset) % CLUSTERS_PER_FAT_SECTOR;
+				if (!(SWAB32(fat[k]) & 0x0fffffff)) {
+					unsigned long c =
+					    nr * CLUSTERS_PER_FAT_SECTOR + k;
+					/*
+					 * Ignore the reserved clusters 0 & 1,
+					 * and also cluster numbers out of
+					 * bounds 
+					 */
+					if (c < 2
+					    || c > fat_bpb.dataclusters + 1)
+						continue;
+					fat_dprintf("startcluster %lu, c %lu\n",
+						    startcluster, c);
+					fat_bpb.fsinfo.nextfree = c;
+					return c;
+				}
+			}
+			offset = 0;
+		}
+	}
+
+	fat_dprintf("startcluster 0\n");
+	return 0;		/* 0 is an illegal cluster number */
+}
+
+static int
+update_fat_entry(unsigned long entry, unsigned long val)
+{
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		int sector = entry / CLUSTERS_PER_FAT16_SECTOR;
+		int offset = entry % CLUSTERS_PER_FAT16_SECTOR;
+		unsigned short *sec;
+
+		val &= 0xFFFF;
+
+		fat_dprintf("entry %x, val %x\n", entry, val);
+
+		if (entry == val)
+			fat_pprintf
+			    ("Creating FAT loop: entry = %lx, val = %lx\n",
+			     entry, val);
+
+		if (entry < 2)
+			fat_pprintf("Updating reserved FAT entry %ld\n", entry);
+
+		sec = cache_fat_sector(sector, true);
+		if (!sec) {
+			fat_eprintf("Could not cache sector %d\n", sector);
+			return -1;
+		}
+
+		if (val) {
+			if (SWAB16(sec[offset]) == 0x0000
+			    && fat_bpb.fsinfo.freecount > 0)
+				fat_bpb.fsinfo.freecount--;
+		} else {
+			if (SWAB16(sec[offset]))
+				fat_bpb.fsinfo.freecount++;
+		}
+
+		fat_dprintf("%d free clusters\n", fat_bpb.fsinfo.freecount);
+		sec[offset] = SWAB16(val);
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{
+		long sector = entry / CLUSTERS_PER_FAT_SECTOR;
+		int offset = entry % CLUSTERS_PER_FAT_SECTOR;
+		unsigned long *sec;
+
+		fat_dprintf("entry %lx, val %lx\n", entry, val);
+
+		if (entry == val)
+			fat_pprintf
+			    ("Creating FAT loop: entry = %lx, val = %lx\n",
+			     entry, val);
+
+		if (entry < 2)
+			fat_pprintf("Updating reserved FAT entry %ld\n", entry);
+
+		sec = cache_fat_sector(sector, true);
+		if (!sec) {
+			fat_eprintf("Could not cache sector %ld\n", sector);
+			return -1;
+		}
+
+		if (val) {
+			if (!(SWAB32(sec[offset]) & 0x0fffffff) &&
+			    fat_bpb.fsinfo.freecount > 0)
+				fat_bpb.fsinfo.freecount--;
+		} else {
+			if (SWAB32(sec[offset]) & 0x0fffffff)
+				fat_bpb.fsinfo.freecount++;
+		}
+
+		fat_dprintf("%ld free clusters\n", fat_bpb.fsinfo.freecount);
+
+		/*
+		 * don't change top 4 bits 
+		 */
+		sec[offset] &= SWAB32(0xf0000000);
+		sec[offset] |= SWAB32(val & 0x0fffffff);
+	}
+
+	return 0;
+}
+
+static long
+read_fat_entry(unsigned long entry)
+{
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		int sector = entry / CLUSTERS_PER_FAT16_SECTOR;
+		int offset = entry % CLUSTERS_PER_FAT16_SECTOR;
+		unsigned short *sec;
+
+		sec = cache_fat_sector(sector, false);
+		if (!sec) {
+			fat_eprintf("Could not cache sector %d\n", sector);
+			return -1;
+		}
+
+		return SWAB16(sec[offset]);
+	} else
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+	{
+		long sector = entry / CLUSTERS_PER_FAT_SECTOR;
+		int offset = entry % CLUSTERS_PER_FAT_SECTOR;
+		unsigned long *sec;
+
+		sec = cache_fat_sector(sector, false);
+		if (!sec) {
+			fat_eprintf("Could not cache sector %ld\n", sector);
+			return -1;
+		}
+
+		return SWAB32(sec[offset]) & 0x0fffffff;
+	}
+}
+
+static long
+get_next_cluster(long cluster)
+{
+	long next_cluster;
+	long eof_mark = FAT_EOF_MARK;
+
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16) {
+		eof_mark &= 0xFFFF;	/* only 16 bit */
+		if (cluster < 0)	/* FAT16 root dir */
+			return cluster + 1;	/* don't use the FAT */
+	}
+#endif
+	next_cluster = read_fat_entry(cluster);
+
+	/*
+	 * is this last cluster in chain? 
+	 */
+	if (next_cluster >= eof_mark)
+		return 0;
+	else
+		return next_cluster;
+}
+
+static int
+update_fsinfo(void)
+{
+	unsigned char fsinfo[SECTOR_SIZE];
+	unsigned long *intptr;
+	int rc;
+
+#ifdef HAVE_FAT16SUPPORT
+	if (fat_bpb.is_fat16)
+		return 0;	/* FAT16 has no FsInfo */
+#endif				/* #ifdef HAVE_FAT16SUPPORT */
+
+	/*
+	 * update fsinfo 
+	 */
+	rc = disk_read(fat_bpb.startsector + fat_bpb.bpb_fsinfo, 1, fsinfo);
+	if (rc < 0) {
+		fat_eprintf("FSInfo read failed (rc = %d)\n", rc);
+		return rc * 10 - 1;
+	}
+	intptr = (long *) &(fsinfo[FSINFO_FREECOUNT]);
+	*intptr = SWAB32(fat_bpb.fsinfo.freecount);
+
+	intptr = (long *) &(fsinfo[FSINFO_NEXTFREE]);
+	*intptr = SWAB32(fat_bpb.fsinfo.nextfree);
+
+	rc = disk_write(fat_bpb.startsector + fat_bpb.bpb_fsinfo, 1, fsinfo);
+	if (rc < 0) {
+		fat_eprintf("FSInfo write failed (rc = %d)\n", rc);
+		return rc * 10 - 2;
+	}
+
+	return 0;
+}
+
+static int
+flush_fat(void)
+{
+	int i;
+	int rc;
+	unsigned char *sec;
+	fat_dprintf("\n");
+
+	for (i = 0; i < FAT_CACHE_SIZE; i++) {
+		struct fat_cache_entry *fce = &fat_cache[i];
+		if (fce->inuse && fce->dirty) {
+			sec = fat_cache_sectors[i];
+			flush_fat_sector(fce, sec);
+		}
+	}
+
+	rc = update_fsinfo();
+	if (rc < 0)
+		return rc * 10 - 3;
+
+	return 0;
+}
+
+static void
+fat_time(unsigned short *date, unsigned short *time, unsigned short *tenth)
+{
+#ifdef HAVE_RTC
+	struct tm *tm = get_time();
+
+	if (date)
+		*date = ((tm->tm_year - 80) << 9) |
+		    ((tm->tm_mon + 1) << 5) | tm->tm_mday;
+
+	if (time)
+		*time =
+		    (tm->tm_hour << 11) | (tm->tm_min << 5) | (tm->tm_sec >> 1);
+
+	if (tenth)
+		*tenth = (tm->tm_sec & 1) * 100;
+#else
+	/*
+	 * non-RTC version returns an increment from the supplied time, or a
+	 * fixed standard time/date if no time given as input 
+	 */
+	bool next_day = false;
+
+	if (time) {
+		if (0 == *time) {
+			/*
+			 * set to 00:15:00 
+			 */
+			*time = (15 << 5);
+		} else {
+			unsigned short mins = (*time >> 5) & 0x003F;
+			unsigned short hours = (*time >> 11) & 0x001F;
+			if ((mins += 10) >= 60) {
+				mins = 0;
+				hours++;
+			}
+			if ((++hours) >= 24) {
+				hours = hours - 24;
+				next_day = true;
+			}
+			*time = (hours << 11) | (mins << 5);
+		}
+	}
+
+	if (date) {
+		if (0 == *date) {
+/* Macros to convert a 2-digit string to a decimal constant. 
+   (YEAR), MONTH and DAY are set by the date command, which outputs
+   DAY as 00..31 and MONTH as 01..12. The leading zero would lead to
+   misinterpretation as an octal constant. */
+
+/* FIXME: Tie into U-Boot's date stuff */
+#define	YEAR	2006
+#define	MONTH	01
+#define	DAY		19
+
+#define S100(x) 1 ## x
+#define C2DIG2DEC(x) (S100(x)-100)
+			/*
+			 * set to build date 
+			 */
+			*date = ((YEAR - 1980) << 9) | (C2DIG2DEC(MONTH) << 5)
+			    | C2DIG2DEC(DAY);
+		} else {
+			unsigned short day = *date & 0x001F;
+			unsigned short month = (*date >> 5) & 0x000F;
+			unsigned short year = (*date >> 9) & 0x007F;
+			if (next_day) {
+				/*
+				 * do a very simple day increment - never go
+				 * above 28 days 
+				 */
+				if (++day > 28) {
+					day = 1;
+					if (++month > 12) {
+						month = 1;
+						year++;
+					}
+				}
+				*date = (year << 9) | (month << 5) | day;
+			}
+		}
+	}
+	if (tenth)
+		*tenth = 0;
+#endif				/* HAVE_RTC */
+}
+
+static int
+write_long_name(struct fat_file *file,
+		unsigned int firstentry,
+		unsigned int numentries,
+		const unsigned char *name,
+		const unsigned char *shortname, bool is_directory)
+{
+	unsigned char buf[SECTOR_SIZE];
+	unsigned char *entry;
+	unsigned int idx = firstentry % DIR_ENTRIES_PER_SECTOR;
+	unsigned int sector = firstentry / DIR_ENTRIES_PER_SECTOR;
+	unsigned int i, j = 0;
+	unsigned char chksum = 0;
+	int nameidx = 0, namelen = strlen(name);
+	int rc;
+
+	fat_dprintf("file %ld, first %d, num %d, name \"%s\"\n",
+		    file->firstcluster, firstentry, numentries, name);
+
+	rc = fat_seek(file, sector);
+	if (rc < 0)
+		return rc * 10 - 1;
+
+	rc = fat_readwrite(file, 1, buf, false);
+	if (rc < 1)
+		return rc * 10 - 2;
+
+	/*
+	 * calculate shortname checksum 
+	 */
+	for (i = 11; i > 0; i--)
+		chksum =
+		    ((chksum & 1) ? 0x80 : 0) + (chksum >> 1) + shortname[j++];
+
+	/*
+	 * calc position of last name segment 
+	 */
+	if (namelen > NAME_BYTES_PER_ENTRY)
+		for (nameidx = 0;
+		     nameidx < (namelen - NAME_BYTES_PER_ENTRY);
+		     nameidx += NAME_BYTES_PER_ENTRY) ;
+
+	for (i = 0; i < numentries; i++) {
+		/*
+		 * new sector? 
+		 */
+		if (idx >= DIR_ENTRIES_PER_SECTOR) {
+			/*
+			 * update current sector 
+			 */
+			rc = fat_seek(file, sector);
+			if (rc < 0)
+				return rc * 10 - 3;
+
+			rc = fat_readwrite(file, 1, buf, true);
+			if (rc < 1)
+				return rc * 10 - 4;
+
+			/*
+			 * read next sector 
+			 */
+			rc = fat_readwrite(file, 1, buf, false);
+			if (rc < 0) {
+				fat_eprintf
+				    ("Failed to write new sector (rc = %d)\n",
+				     rc);
+				return rc * 10 - 5;
+			}
+			if (rc == 0)
+				/*
+				 * end of dir 
+				 */
+				memset(buf, 0, sizeof buf);
+
+			sector++;
+			idx = 0;
+		}
+
+		entry = buf + idx * DIR_ENTRY_SIZE;
+
+		/*
+		 * verify this entry is free 
+		 */
+		if (entry[0] && entry[0] != 0xe5)
+			/*
+			 * FIXME: Should we halt instead of plowing on? 
+			 */
+			fat_pprintf("Dir entry %d in sector %x is not free! "
+				    "%02x %02x %02x %02x",
+				    idx, sector, entry[0], entry[1], entry[2],
+				    entry[3]);
+
+		memset(entry, 0, DIR_ENTRY_SIZE);
+		if (i + 1 < numentries) {
+			/*
+			 * longname entry 
+			 */
+			int k, l = nameidx;
+
+			entry[FATLONG_ORDER] = numentries - i - 1;
+			if (i == 0) {
+				/*
+				 * mark this as last long entry 
+				 */
+				entry[FATLONG_ORDER] |= 0x40;
+
+				/*
+				 * pad name with 0xffff 
+				 */
+				for (k = 1; k < 12; k++)
+					entry[k] = 0xff;
+				for (k = 14; k < 26; k++)
+					entry[k] = 0xff;
+				for (k = 28; k < 32; k++)
+					entry[k] = 0xff;
+			};
+
+			/*
+			 * set name 
+			 */
+			for (k = 0; k < 5 && l <= namelen; k++) {
+				entry[k * 2 + 1] = name[l++];
+				entry[k * 2 + 2] = 0;
+			}
+
+			for (k = 0; k < 6 && l <= namelen; k++) {
+				entry[k * 2 + 14] = name[l++];
+				entry[k * 2 + 15] = 0;
+			}
+
+			for (k = 0; k < 2 && l <= namelen; k++) {
+				entry[k * 2 + 28] = name[l++];
+				entry[k * 2 + 29] = 0;
+			}
+
+			entry[FATDIR_ATTR] = FAT_ATTR_LONG_NAME;
+			entry[FATDIR_FSTCLUSLO] = 0;
+			entry[FATLONG_TYPE] = 0;
+			entry[FATLONG_CHKSUM] = chksum;
+			fat_dprintf("longname entry %d = \"%s\"\n", idx,
+				    name + nameidx);
+		} else {
+			/*
+			 * shortname entry 
+			 */
+			unsigned short date = 0, time = 0, tenth = 0;
+			fat_dprintf("shortname entry = \"%s\"\n", shortname);
+			strncpy(entry + FATDIR_NAME, shortname, 11);
+			entry[FATDIR_ATTR] =
+			    is_directory ? FAT_ATTR_DIRECTORY : 0;
+			entry[FATDIR_NTRES] = 0;
+
+			fat_time(&date, &time, &tenth);
+			entry[FATDIR_CRTTIMETENTH] = tenth;
+			*(unsigned short *) (entry + FATDIR_CRTTIME) =
+			    SWAB16(time);
+			*(unsigned short *) (entry + FATDIR_WRTTIME) =
+			    SWAB16(time);
+			*(unsigned short *) (entry + FATDIR_CRTDATE) =
+			    SWAB16(date);
+			*(unsigned short *) (entry + FATDIR_WRTDATE) =
+			    SWAB16(date);
+			*(unsigned short *) (entry + FATDIR_LSTACCDATE) =
+			    SWAB16(date);
+		}
+		idx++;
+		nameidx -= NAME_BYTES_PER_ENTRY;
+	}
+
+	/*
+	 * update last sector 
+	 */
+	rc = fat_seek(file, sector);
+	if (rc < 0)
+		return rc * 10 - 6;
+
+	rc = fat_readwrite(file, 1, buf, true);
+	if (rc < 1)
+		return rc * 10 - 7;
+
+	return 0;
+}
+
+static int
+fat_checkname(const unsigned char *newname)
+{
+	/*
+	 * More sanity checks are probably needed 
+	 */
+	if (newname[strlen(newname) - 1] == '.') {
+		return -1;
+	}
+
+	return 0;
+}
+
+static int
+add_dir_entry(struct fat_dir *dir,
+	      struct fat_file *file,
+	      const char *name, bool is_directory, bool dotdir)
+{
+	unsigned char buf[SECTOR_SIZE];
+	unsigned char shortname[12];
+	int rc;
+	unsigned int sector;
+	bool done = false;
+	int entries_needed, entries_found = 0;
+	int firstentry;
+
+	fat_dprintf("name \"%s\", firstcluster %ld\n",
+		    name, file->firstcluster);
+
+	/*
+	 * Don't check dotdirs name for validity 
+	 */
+	if (dotdir == false) {
+		rc = fat_checkname(name);
+		if (rc < 0) {
+			/*
+			 * filename is invalid 
+			 */
+			return rc * 10 - 1;
+		}
+	}
+
+	/*
+	 * The "." and ".." directory entries must not be long names 
+	 */
+	if (dotdir) {
+		int i;
+		strncpy(shortname, name, 12);
+		for (i = strlen(shortname); i < 12; i++)
+			shortname[i] = ' ';
+
+		entries_needed = 1;
+	} else {
+		create_dos_name(name, shortname);
+
+		/*
+		 * one dir entry needed for every 13 bytes of filename, plus
+		 * one entry for the short name 
+		 */
+		entries_needed = (strlen(name) + (NAME_BYTES_PER_ENTRY - 1))
+		    / NAME_BYTES_PER_ENTRY + 1;
+	}
+
+      restart:
+	firstentry = -1;
+
+	rc = fat_seek(&dir->file, 0);
+	if (rc < 0)
+		return rc * 10 - 2;
+
+	/*
+	 * step 1: search for free entries and check for duplicate shortname 
+	 */
+	for (sector = 0; !done; sector++) {
+		unsigned int i;
+
+		rc = fat_readwrite(&dir->file, 1, buf, false);
+		if (rc < 0) {
+			fat_eprintf("Could not read dir (rc = %d)\n", rc);
+			return rc * 10 - 3;
+		}
+
+		if (rc == 0) {	/* current end of dir reached */
+			fat_dprintf("End of dir on cluster boundary\n");
+			break;
+		}
+
+		/*
+		 * look for free slots 
+		 */
+		for (i = 0; i < DIR_ENTRIES_PER_SECTOR; i++) {
+			switch (buf[i * DIR_ENTRY_SIZE]) {
+			case 0:
+				entries_found += DIR_ENTRIES_PER_SECTOR - i;
+				fat_dprintf("Found end of dir %d\n",
+					    sector * DIR_ENTRIES_PER_SECTOR +
+					    i);
+				i = DIR_ENTRIES_PER_SECTOR - 1;
+				done = true;
+				break;
+
+			case 0xe5:
+				entries_found++;
+				fat_dprintf("Found free entry %d (%d/%d)\n",
+					    sector * DIR_ENTRIES_PER_SECTOR + i,
+					    entries_found, entries_needed);
+				break;
+
+			default:
+				entries_found = 0;
+
+				/*
+				 * check that our intended shortname doesn't
+				 * already exist 
+				 */
+				if (!strncmp
+				    (shortname, buf + i * DIR_ENTRY_SIZE, 12)) {
+					/*
+					 * shortname exists already, make a new 
+					 * one 
+					 */
+					randomize_dos_name(shortname);
+					fat_dprintf
+					    ("Duplicate shortname, changing to \"%s\"\n",
+					     shortname);
+
+					/*
+					 * name has changed, we need to restart 
+					 * search 
+					 */
+					goto restart;
+				}
+				break;
+			}
+			if (firstentry < 0 && (entries_found >= entries_needed))
+				firstentry =
+				    sector * DIR_ENTRIES_PER_SECTOR + i + 1 -
+				    entries_found;
+		}
+	}
+
+	/*
+	 * step 2: extend the dir if necessary 
+	 */
+	if (firstentry < 0) {
+		fat_dprintf("Adding new sector(s) to dir\n");
+		rc = fat_seek(&dir->file, sector);
+		if (rc < 0)
+			return rc * 10 - 4;
+		memset(buf, 0, sizeof buf);
+
+		/*
+		 * we must clear whole clusters 
+		 */
+		for (; (entries_found < entries_needed) ||
+		     (dir->file.sectornum < (int) fat_bpb.bpb_secperclus);
+		     sector++) {
+			if (sector >= (65536 / DIR_ENTRIES_PER_SECTOR))
+				return -5;	/* dir too large -- FAT
+						 * specification */
+
+			rc = fat_readwrite(&dir->file, 1, buf, true);
+			if (rc < 1)	/* No more room or something went wrong 
+					 */
+				return rc * 10 - 6;
+
+			entries_found += DIR_ENTRIES_PER_SECTOR;
+		}
+
+		firstentry = sector * DIR_ENTRIES_PER_SECTOR - entries_found;
+	}
+
+	/*
+	 * step 3: add entry 
+	 */
+	sector = firstentry / DIR_ENTRIES_PER_SECTOR;
+	fat_dprintf("Adding longname to entry %d in sector %d\n",
+		    firstentry, sector);
+
+	rc = write_long_name(&dir->file, firstentry,
+			     entries_needed, name, shortname, is_directory);
+	if (rc < 0)
+		return rc * 10 - 7;
+
+	/*
+	 * remember where the shortname dir entry is located 
+	 */
+	file->direntry = firstentry + entries_needed - 1;
+	file->direntries = entries_needed;
+	file->dircluster = dir->file.firstcluster;
+	fat_dprintf("Added new dir entry %d, using %d slots\n",
+		    file->direntry, file->direntries);
+	return 0;
+}
+
+unsigned char
+char2dos(unsigned char c)
+{
+	switch (c) {
+	case 0x22:
+	case 0x2a:
+	case 0x2b:
+	case 0x2c:
+	case 0x2e:
+	case 0x3a:
+	case 0x3b:
+	case 0x3c:
+	case 0x3d:
+	case 0x3e:
+	case 0x3f:
+	case 0x5b:
+	case 0x5c:
+	case 0x5d:
+	case 0x7c:
+		/*
+		 * Illegal char, replace 
+		 */
+		c = '_';
+		break;
+
+	default:
+		if (c <= 0x20)
+			c = 0;	/* Illegal char, remove */
+		else
+			c = __toupper(c);
+		break;
+	}
+
+	return c;
+}
+
+static void
+create_dos_name(const unsigned char *name, unsigned char *newname)
+{
+	int i;
+	unsigned char *ext;
+
+	/*
+	 * Find extension part 
+	 */
+	ext = strrchr(name, '.');
+	if (ext == name)	/* handle .dotnames */
+		ext = NULL;
+
+	/*
+	 * Name part 
+	 */
+	for (i = 0; *name && (!ext || name < ext) && (i < 8); name++) {
+		unsigned char c = char2dos(*name);
+		if (c)
+			newname[i++] = c;
+	}
+
+	/*
+	 * Pad both name and extension 
+	 */
+	while (i < 11)
+		newname[i++] = ' ';
+
+	if (newname[0] == 0xe5)	/* Special kanji character */
+		newname[0] = 0x05;
+
+	if (ext) {		/* Extension part */
+		ext++;
+		for (i = 8; *ext && (i < 11); ext++) {
+			unsigned char c = char2dos(*ext);
+			if (c)
+				newname[i++] = c;
+		}
+	}
+}
+
+/*
+ * This function is hacked to always start with a fixed number and keep
+ * incrementing it until the caller is satisfied with the name.
+ * FIXME: add a real randomizer?
+ */
+static void
+randomize_dos_name(unsigned char *name)
+{
+	static unsigned int random = 0;
+	int i;
+	unsigned char buf[5];
+
+	/*
+	 * snprintf(buf, sizeof buf, "%04X", (unsigned)rand() & 0xffff); 
+	 */
+	sprintf(buf, "%04X", (random++ & 0xffff));
+
+	for (i = 0; (i < 4) && (name[i] != ' '); i++) ;
+	/*
+	 * account for possible shortname length < 4 
+	 */
+	memcpy(&name[i], buf, 4);
+}
+
+static int
+update_short_entry(struct fat_file *file, long size, int attr)
+{
+	unsigned char buf[SECTOR_SIZE];
+	int sector = file->direntry / DIR_ENTRIES_PER_SECTOR;
+	unsigned char *entry =
+	    buf + DIR_ENTRY_SIZE * (file->direntry % DIR_ENTRIES_PER_SECTOR);
+	unsigned long *sizeptr;
+	unsigned short *clusptr;
+	struct fat_file dir;
+	int rc;
+
+	fat_dprintf("cluster %lu, entry %d, size %ld\n",
+		    file->firstcluster, file->direntry, size);
+
+	/*
+	 * create a temporary file handle for the dir holding this file 
+	 */
+	rc = fat_open(file->dircluster, &dir, NULL);
+	if (rc < 0)
+		return rc * 10 - 1;
+
+	rc = fat_seek(&dir, sector);
+	if (rc < 0)
+		return rc * 10 - 2;
+
+	rc = fat_readwrite(&dir, 1, buf, false);
+	if (rc < 1)
+		return rc * 10 - 3;
+
+	if (!entry[0] || entry[0] == 0xe5)
+		fat_pprintf("Updating size on empty dir entry %d\n",
+			    file->direntry);
+
+	entry[FATDIR_ATTR] = attr & 0xFF;
+
+	clusptr = (short *) (entry + FATDIR_FSTCLUSHI);
+	*clusptr = SWAB16(file->firstcluster >> 16);
+
+	clusptr = (short *) (entry + FATDIR_FSTCLUSLO);
+	*clusptr = SWAB16(file->firstcluster & 0xffff);
+
+	sizeptr = (long *) (entry + FATDIR_FILESIZE);
+	*sizeptr = SWAB32(size);
+
+	{
+#ifdef HAVE_RTC
+		unsigned short time = 0;
+		unsigned short date = 0;
+#else
+		/*
+		 * get old time to increment from 
+		 */
+		unsigned short time =
+		    SWAB16(*(unsigned short *) (entry + FATDIR_WRTTIME));
+		unsigned short date =
+		    SWAB16(*(unsigned short *) (entry + FATDIR_WRTDATE));
+#endif
+		fat_time(&date, &time, NULL);
+		*(unsigned short *) (entry + FATDIR_WRTTIME) = SWAB16(time);
+		*(unsigned short *) (entry + FATDIR_WRTDATE) = SWAB16(date);
+		*(unsigned short *) (entry + FATDIR_LSTACCDATE) = SWAB16(date);
+	}
+
+	rc = fat_seek(&dir, sector);
+	if (rc < 0)
+		return rc * 10 - 4;
+
+	rc = fat_readwrite(&dir, 1, buf, true);
+	if (rc < 1)
+		return rc * 10 - 5;
+
+	return 0;
+}
+
+static int
+parse_direntry(struct fat_direntry *de, const unsigned char *buf)
+{
+	int i = 0, j = 0;
+	unsigned char c;
+	bool lowercase;
+
+	memset(de, 0, sizeof (struct fat_direntry));
+	de->attr = buf[FATDIR_ATTR];
+	de->crttimetenth = buf[FATDIR_CRTTIMETENTH];
+	de->crtdate = BYTES2INT16(buf, FATDIR_CRTDATE);
+	de->crttime = BYTES2INT16(buf, FATDIR_CRTTIME);
+	de->wrtdate = BYTES2INT16(buf, FATDIR_WRTDATE);
+	de->wrttime = BYTES2INT16(buf, FATDIR_WRTTIME);
+	de->filesize = BYTES2INT32(buf, FATDIR_FILESIZE);
+	de->firstcluster =
+	    ((long) (unsigned) BYTES2INT16(buf, FATDIR_FSTCLUSLO)) |
+	    ((long) (unsigned) BYTES2INT16(buf, FATDIR_FSTCLUSHI) << 16);
+	/*
+	 * The double cast is to prevent a sign-extension to be done on
+	 * CalmRISC16 (the result of the shift is always considered signed) 
+	 */
+
+	/*
+	 * fix the name 
+	 */
+	lowercase = (buf[FATDIR_NTRES] & FAT_NTRES_LC_NAME);
+	c = buf[FATDIR_NAME];
+	if (c == 0x05)		/* special kanji char */
+		c = 0xe5;
+	i = 0;
+	while (c != ' ') {
+		de->name[j++] = lowercase ? __tolower(c) : c;
+		if (++i >= 8)
+			break;
+		c = buf[FATDIR_NAME + i];
+	}
+
+	if (buf[FATDIR_NAME + 8] != ' ') {
+		lowercase = (buf[FATDIR_NTRES] & FAT_NTRES_LC_EXT);
+		de->name[j++] = '.';
+		for (i = 8; (i < 11) && ((c = buf[FATDIR_NAME + i]) != ' ');
+		     i++)
+			de->name[j++] = lowercase ? __tolower(c) : c;
+	}
+	return 1;
+}
+
+int
+fat_open(long startcluster, struct fat_file *file, const struct fat_dir *dir)
+{
+	file->firstcluster = startcluster;
+	file->lastcluster = startcluster;
+	file->lastsector = 0;
+	file->clusternum = 0;
+	file->sectornum = 0;
+	file->eof = false;
+
+	/*
+	 * remember where the file's dir entry is located 
+	 */
+	if (dir) {
+		file->direntry = dir->entry - 1;
+		file->direntries = dir->entrycount;
+		file->dircluster = dir->file.firstcluster;
+	}
+
+	fat_dprintf("startcluster %ld, direntry %d\n", startcluster,
+		    file->direntry);
+	return 0;
+}
+
+int
+fat_create_file(const char *name, struct fat_file *file, struct fat_dir *dir)
+{
+	int rc;
+
+	fat_dprintf("name \"%s\", file ptr %lx, dir ptr %lx\n",
+		    name, (long) file, (long) dir);
+	rc = add_dir_entry(dir, file, name, false, false);
+	if (!rc) {
+		file->firstcluster = 0;
+		file->lastcluster = 0;
+		file->lastsector = 0;
+		file->clusternum = 0;
+		file->sectornum = 0;
+		file->eof = false;
+	}
+
+	return rc;
+}
+
+int
+fat_create_dir(const char *name, struct fat_dir *newdir, struct fat_dir *dir)
+{
+	unsigned char buf[SECTOR_SIZE];
+	int i;
+	long sector;
+	int rc;
+	struct fat_file dummyfile;
+
+	fat_dprintf("name \"%s\", newdir ptr %lx, dir ptr %lx\n",
+		    name, (long) newdir, (long) dir);
+
+	memset(newdir, 0, sizeof (struct fat_dir));
+	memset(&dummyfile, 0, sizeof (struct fat_file));
+
+	/*
+	 * First, add the entry in the parent directory 
+	 */
+	rc = add_dir_entry(dir, &newdir->file, name, true, false);
+	if (rc < 0)
+		return rc * 10 - 1;
+
+	/*
+	 * Allocate a new cluster for the directory 
+	 */
+	newdir->file.firstcluster = find_free_cluster(fat_bpb.fsinfo.nextfree);
+	if (newdir->file.firstcluster == 0)
+		return -1;
+
+	update_fat_entry(newdir->file.firstcluster, FAT_EOF_MARK);
+
+	/*
+	 * Clear the entire cluster 
+	 */
+	memset(buf, 0, sizeof buf);
+	sector = cluster2sec(newdir->file.firstcluster);
+	for (i = 0; i < (int) fat_bpb.bpb_secperclus; i++) {
+		rc = transfer(sector + i, 1, buf, true);
+		if (rc < 0)
+			return rc * 10 - 2;
+	}
+
+	/*
+	 * Then add the "." entry 
+	 */
+	rc = add_dir_entry(newdir, &dummyfile, ".", true, true);
+	if (rc < 0)
+		return rc * 10 - 3;
+	dummyfile.firstcluster = newdir->file.firstcluster;
+	update_short_entry(&dummyfile, 0, FAT_ATTR_DIRECTORY);
+
+	/*
+	 * and the ".." entry 
+	 */
+	rc = add_dir_entry(newdir, &dummyfile, "..", true, true);
+	if (rc < 0)
+		return rc * 10 - 4;
+
+	/*
+	 * The root cluster is cluster 0 in the ".." entry 
+	 */
+	if (dir->file.firstcluster == fat_bpb.bpb_rootclus)
+		dummyfile.firstcluster = 0;
+	else
+		dummyfile.firstcluster = dir->file.firstcluster;
+	update_short_entry(&dummyfile, 0, FAT_ATTR_DIRECTORY);
+
+	/*
+	 * Set the firstcluster field in the direntry 
+	 */
+	update_short_entry(&newdir->file, 0, FAT_ATTR_DIRECTORY);
+
+	rc = flush_fat();
+	if (rc < 0)
+		return rc * 10 - 5;
+
+	return rc;
+}
+
+int
+fat_truncate(const struct fat_file *file)
+{
+	/*
+	 * truncate trailing clusters 
+	 */
+	long next;
+	long last = file->lastcluster;
+
+	fat_dprintf("firstcluster %ld, lastcluster %ld\n",
+		    file->firstcluster, last);
+
+	for (last = get_next_cluster(last); last; last = next) {
+		next = get_next_cluster(last);
+		update_fat_entry(last, 0);
+	}
+
+	if (file->lastcluster)
+		update_fat_entry(file->lastcluster, FAT_EOF_MARK);
+
+	return 0;
+}
+
+int
+fat_closewrite(struct fat_file *file, long size, int attr)
+{
+	int rc;
+
+	fat_dprintf("size %ld\n", size);
+	if (!size) {
+		/*
+		 * empty file 
+		 */
+		if (file->firstcluster) {
+			update_fat_entry(file->firstcluster, 0);
+			file->firstcluster = 0;
+		}
+	}
+
+	if (file->dircluster) {
+		rc = update_short_entry(file, size, attr);
+		if (rc < 0)
+			return rc * 10 - 1;
+	}
+
+	flush_fat();
+
+#ifdef TEST_FAT
+	if (file->firstcluster) {
+		/*
+		 * debug 
+		 */
+		long count = 0;
+		long len;
+		long next;
+		for (next = file->firstcluster; next;
+		     next = get_next_cluster(next)) {
+			fat_dprintf("cluster %ld, next %ld\n", count, next);
+			count++;
+		}
+
+		len = count * fat_bpb.bpb_secperclus * SECTOR_SIZE;
+		fat_dprintf("File is %ld clusters (chainlen %ld, size %ld)\n",
+			    count, len, size);
+		if (len > size + fat_bpb.bpb_secperclus * SECTOR_SIZE)
+			fat_pprintf("Cluster chain is too long\n");
+
+		if (len < size)
+			fat_pprintf("Cluster chain is too short\n");
+	}
+#endif
+
+	return 0;
+}
+
+static int
+free_direntries(struct fat_file *file)
+{
+	unsigned char buf[SECTOR_SIZE];
+	struct fat_file dir;
+	int numentries = file->direntries;
+	unsigned int entry = file->direntry - numentries + 1;
+	unsigned int sector = entry / DIR_ENTRIES_PER_SECTOR;
+	int i;
+	int rc;
+
+	/*
+	 * create a temporary file handle for the dir holding this file 
+	 */
+	rc = fat_open(file->dircluster, &dir, NULL);
+	if (rc < 0)
+		return rc * 10 - 1;
+
+	rc = fat_seek(&dir, sector);
+	if (rc < 0)
+		return rc * 10 - 2;
+
+	rc = fat_readwrite(&dir, 1, buf, false);
+	if (rc < 1)
+		return rc * 10 - 3;
+
+	for (i = 0; i < numentries; i++) {
+		fat_dprintf("Clearing dir entry %d (%d/%d)\n",
+			    entry, i + 1, numentries);
+		buf[(entry % DIR_ENTRIES_PER_SECTOR) * DIR_ENTRY_SIZE] = 0xe5;
+		entry++;
+
+		if ((entry % DIR_ENTRIES_PER_SECTOR) == 0) {
+			/*
+			 * flush this sector 
+			 */
+			rc = fat_seek(&dir, sector);
+			if (rc < 0)
+				return rc * 10 - 4;
+
+			rc = fat_readwrite(&dir, 1, buf, true);
+			if (rc < 1)
+				return rc * 10 - 5;
+
+			if (i + 1 < numentries) {
+				/*
+				 * read next sector 
+				 */
+				rc = fat_readwrite(&dir, 1, buf, false);
+				if (rc < 1)
+					return rc * 10 - 6;
+			}
+			sector++;
+		}
+	}
+
+	if (entry % DIR_ENTRIES_PER_SECTOR) {
+		/*
+		 * flush this sector 
+		 */
+		rc = fat_seek(&dir, sector);
+		if (rc < 0)
+			return rc * 10 - 7;
+
+		rc = fat_readwrite(&dir, 1, buf, true);
+		if (rc < 1)
+			return rc * 10 - 8;
+	}
+
+	return 0;
+}
+
+int
+fat_remove(struct fat_file *file)
+{
+	long next, last = file->firstcluster;
+	int rc;
+
+	fat_dprintf("last %ld\n", last);
+
+	while (last) {
+		next = get_next_cluster(last);
+		update_fat_entry(last, 0);
+		last = next;
+	}
+
+	if (file->dircluster) {
+		rc = free_direntries(file);
+		if (rc < 0)
+			return rc * 10 - 1;
+	}
+
+	file->firstcluster = 0;
+	file->dircluster = 0;
+
+	rc = flush_fat();
+	if (rc < 0)
+		return rc * 10 - 2;
+
+	return 0;
+}
+
+int
+fat_rename(struct fat_file *file,
+	   struct fat_dir *dir,
+	   const unsigned char *newname, long size, int attr)
+{
+	int rc;
+	struct fat_dir olddir;
+	struct fat_file newfile = *file;
+
+	if (!file->dircluster) {
+		fat_eprintf("File has no dir cluster!\n");
+		return -2;
+	}
+
+	/*
+	 * create a temporary file handle 
+	 */
+	rc = fat_opendir(&olddir, file->dircluster, NULL);
+	if (rc < 0)
+		return rc * 10 - 3;
+
+	/*
+	 * create new name 
+	 */
+	rc = add_dir_entry(dir, &newfile, newname, false, false);
+	if (rc < 0)
+		return rc * 10 - 4;
+
+	/*
+	 * write size and cluster link 
+	 */
+	rc = update_short_entry(&newfile, size, attr);
+	if (rc < 0)
+		return rc * 10 - 5;
+
+	/*
+	 * remove old name 
+	 */
+	rc = free_direntries(file);
+	if (rc < 0)
+		return rc * 10 - 6;
+
+	rc = flush_fat();
+	if (rc < 0)
+		return rc * 10 - 7;
+
+	return 0;
+}
+
+static long
+next_write_cluster(struct fat_file *file, long oldcluster, long *newsector)
+{
+	long cluster = 0;
+	long sector;
+
+/*
+ * Enable this for lots of verbosity.
+ */
+#if 0
+	fat_dprintf("firstcluster %ld, lastcluster %ld\n",
+		    file->firstcluster, oldcluster);
+#endif
+
+	if (oldcluster)
+		cluster = get_next_cluster(oldcluster);
+
+	if (!cluster) {
+		if (oldcluster > 0)
+			cluster = find_free_cluster(oldcluster + 1);
+		else if (oldcluster == 0)
+			cluster = find_free_cluster(fat_bpb.fsinfo.nextfree);
+#ifdef HAVE_FAT16SUPPORT
+		else		/* negative, pseudo-cluster of the root dir */
+			return 0;	/* impossible to append something to
+					 * the root */
+#endif
+
+		if (cluster) {
+			if (oldcluster)
+				update_fat_entry(oldcluster, cluster);
+			else
+				file->firstcluster = cluster;
+			update_fat_entry(cluster, FAT_EOF_MARK);
+		} else {
+#ifdef TEST_FAT
+			if (fat_bpb.fsinfo.freecount > 0)
+				fat_pprintf("There is free space, but "
+					    "find_free_cluster() didn't find it!\n");
+#endif
+			fat_eprintf("Disk is full!\n");
+			return 0;
+		}
+	}
+	sector = cluster2sec(cluster);
+	if (sector < 0)
+		return 0;
+
+	*newsector = sector;
+	return cluster;
+}
+
+static int
+transfer(unsigned long start, long count, char *buf, bool write)
+{
+	int rc;
+
+	fat_dprintf("%s, start sector %lu, count %ld\n",
+		    write ? "Write" : "Read", start + fat_bpb.startsector,
+		    count);
+	if (write) {
+		unsigned long firstallowed;
+#ifdef HAVE_FAT16SUPPORT
+		if (fat_bpb.is_fat16)
+			firstallowed = fat_bpb.rootdirsector;
+		else
+#endif
+			firstallowed = fat_bpb.firstdatasector;
+
+		if (start < firstallowed) {
+			fat_pprintf("Write %ld before data\n",
+				    firstallowed - start);
+			return -1;
+		}
+
+		if (start + count > fat_bpb.totalsectors) {
+			fat_pprintf("Write %ld after data\n",
+				    start + count - fat_bpb.totalsectors);
+			return -1;
+		}
+
+		rc = disk_write(start + fat_bpb.startsector, count, buf);
+	} else {
+		rc = disk_read(start + fat_bpb.startsector, count, buf);
+	}
+
+	if (rc < 0) {
+		fat_eprintf
+		    ("%s failed: start sector %lu, count %ld (rc = %d)\n",
+		     write ? "Write" : "Read", start, count, rc);
+		return rc;
+	}
+
+	return 0;
+}
+
+long
+fat_readwrite(struct fat_file *file, long sectorcount, void *buf, bool write)
+{
+	long cluster = file->lastcluster;
+	long sector = file->lastsector;
+	long clusternum = file->clusternum;
+	long numsec = file->sectornum;
+	bool eof = file->eof;
+	long first = 0, last = 0;
+	long i;
+	int rc;
+
+	fat_dprintf("%s, firstcluster %ld, count %ld, buf 0x%lx, sector %ld, "
+		    "numsec %ld, %s\n",
+		    write ? "Write" : "Read", file->firstcluster,
+		    sectorcount, (long) buf, sector, numsec,
+		    eof ? "eof" : "noeof");
+
+	if (eof && !write)
+		return 0;
+
+	/*
+	 * find sequential sectors and write them all at once 
+	 */
+	for (i = 0; (i < sectorcount) && (sector > -1); i++) {
+		numsec++;
+		if (numsec > (long) fat_bpb.bpb_secperclus || !cluster) {
+			long oldcluster = cluster;
+			if (write)
+				cluster =
+				    next_write_cluster(file, cluster, &sector);
+			else {
+				cluster = get_next_cluster(cluster);
+				sector = cluster2sec(cluster);
+			}
+
+			clusternum++;
+			numsec = 1;
+
+			if (!cluster) {
+				eof = true;
+				if (write) {
+					/*
+					 * remember last cluster, in case we
+					 * want to append to the file 
+					 */
+					cluster = oldcluster;
+					clusternum--;
+					i = -1;	/* Error code */
+					break;
+				}
+			} else
+				eof = false;
+		} else {
+			if (sector) {
+				sector++;
+			} else {
+				/*
+				 * look up first sector of file 
+				 */
+				sector = cluster2sec(file->firstcluster);
+				numsec = 1;
+#ifdef HAVE_FAT16SUPPORT
+				if (file->firstcluster < 0) {	/* FAT16 root
+								 * dir */
+					sector += fat_bpb.rootdiroffset;
+					numsec += fat_bpb.rootdiroffset;
+				}
+#endif
+			}
+		}
+
+		if (!first)
+			first = sector;
+
+		/*
+		 * The original rockbox FAT code had a hard-coded limit of
+		 * 256 sectors max per transfer.  We allow more than that
+		 * now, but BE SURE that your disk driver is able to deal
+		 * with a transfer request of more than 256
+		 * sectors if you do not use the default!
+		 */
+		if (((sector != first) && (sector != last + 1)) ||	/* not
+									 * sequential 
+									 */
+		    (last - first + 1 == CFG_ROCKBOX_FAT_MAX_SECS_PER_XFER)) {
+			long count = last - first + 1;
+			rc = transfer(first, count, buf, write);
+			if (rc < 0)
+				return rc * 10 - 1;
+
+			buf = (char *) buf + count * SECTOR_SIZE;
+			first = sector;
+		}
+
+		if ((i == sectorcount - 1) &&	/* last sector requested */
+		    (!eof)) {
+			long count = sector - first + 1;
+			rc = transfer(first, count, buf, write);
+			if (rc < 0)
+				return rc * 10 - 2;
+		}
+
+		last = sector;
+	}
+
+	file->lastcluster = cluster;
+	file->lastsector = sector;
+	file->clusternum = clusternum;
+	file->sectornum = numsec;
+	file->eof = eof;
+
+	/*
+	 * if eof, don't report last block as read/written 
+	 */
+	if (eof)
+		i--;
+
+	fat_dprintf("Sectors %s = %ld\n", write ? "written" : "read", i);
+	return i;
+}
+
+int
+fat_seek(struct fat_file *file, unsigned long seeksector)
+{
+	long clusternum = 0, numclusters = 0, sectornum = 0, sector = 0;
+	long cluster = file->firstcluster;
+	long i;
+
+#ifdef HAVE_FAT16SUPPORT
+	if (cluster < 0)	/* FAT16 root dir */
+		seeksector += fat_bpb.rootdiroffset;
+#endif
+
+	file->eof = false;
+	if (seeksector) {
+		/*
+		 * we need to find the sector BEFORE the requested, since the
+		 * file struct stores the last accessed sector 
+		 */
+		seeksector--;
+		numclusters = clusternum = seeksector / fat_bpb.bpb_secperclus;
+		sectornum = seeksector % fat_bpb.bpb_secperclus;
+
+		if (file->clusternum && clusternum >= file->clusternum) {
+			cluster = file->lastcluster;
+			numclusters -= file->clusternum;
+		}
+
+		for (i = 0; i < numclusters; i++) {
+			cluster = get_next_cluster(cluster);
+			if (!cluster) {
+				fat_eprintf
+				    ("Attempt to seek beyond end of file "
+				     "(sector %ld, cluster %ld)\n", seeksector,
+				     i);
+				return -1;
+			}
+		}
+
+		sector = cluster2sec(cluster) + sectornum;
+	} else {
+		sectornum = -1;
+	}
+
+	fat_dprintf("firstcluser %ld, seeksector %lu, cluster %ld, "
+		    "sector %ld, sectornum %ld\n",
+		    file->firstcluster, seeksector, cluster, sector, sectornum);
+
+	file->lastcluster = cluster;
+	file->lastsector = sector;
+	file->clusternum = clusternum;
+	file->sectornum = sectornum + 1;
+	return 0;
+}
+
+int
+fat_opendir(struct fat_dir *dir, unsigned long startcluster,
+	    const struct fat_dir *parent_dir)
+{
+	int rc;
+
+	dir->entry = 0;
+	dir->sector = 0;
+
+	if (startcluster == 0)
+		startcluster = fat_bpb.bpb_rootclus;
+
+	rc = fat_open(startcluster, &dir->file, parent_dir);
+	if (rc) {
+		fat_eprintf("Could not open dir (rc = %d)\n", rc);
+		return rc * 10 - 1;
+	}
+
+	return 0;
+}
+
+/* convert from unicode to a single-byte charset */
+static void
+unicode2iso(const unsigned char *unicode, unsigned char *iso, int count)
+{
+	int i;
+
+	for (i = 0; i < count; i++) {
+		int x = i * 2;
+		switch (unicode[x + 1]) {
+		case 0x01:	/* latin extended. convert to ISO 8859-2 */
+		case 0x02:
+			iso[i] = unicode2iso8859_2[unicode[x]];
+			break;
+
+		case 0x03:	/* greek, convert to ISO 8859-7 */
+			iso[i] = unicode[x] + 0x30;
+			break;
+
+			/*
+			 * Sergei says most russians use Win1251, so we will
+			 * too. Win1251 differs from ISO 8859-5 by an offset of 
+			 * 0x10. 
+			 */
+		case 0x04:	/* cyrillic, convert to Win1251 */
+			switch (unicode[x]) {
+			case 1:
+				iso[i] = 168;
+				break;
+
+			case 81:
+				iso[i] = 184;
+				break;
+
+			default:
+				iso[i] = unicode[x] + 0xb0;	/* 0xa0 for ISO 
+								 * 8859-5 */
+				break;
+			}
+			break;
+
+		case 0x05:	/* hebrew, convert to ISO 8859-8 */
+			iso[i] = unicode[x] + 0x10;
+			break;
+
+		case 0x06:	/* arabic, convert to ISO 8859-6 */
+		case 0x0e:	/* thai, convert to ISO 8859-11 */
+			iso[i] = unicode[x] + 0xa0;
+			break;
+
+		default:
+			iso[i] = unicode[x];
+			break;
+		}
+	}
+}
+
+int
+fat_getnext(struct fat_dir *dir, struct fat_direntry *entry)
+{
+	bool done = false;
+	int i;
+	int rc;
+	unsigned char firstbyte;
+	int longarray[20];
+	int longs = 0;
+	int sectoridx = 0;
+	unsigned char *cached_buf = dir->sectorcache[0];
+
+	dir->entrycount = 0;
+
+	while (!done) {
+		if (!(dir->entry % DIR_ENTRIES_PER_SECTOR) || !dir->sector) {
+			rc = fat_readwrite(&dir->file, 1, cached_buf, false);
+			if (rc == 0) {
+				/*
+				 * eof 
+				 */
+				entry->name[0] = 0;
+				break;
+			}
+
+			if (rc < 0) {
+				fat_eprintf("Could not read dir (rc = %d)\n",
+					    rc);
+				return rc * 10 - 1;
+			}
+
+			dir->sector = dir->file.lastsector;
+		}
+
+		for (i = dir->entry % DIR_ENTRIES_PER_SECTOR;
+		     i < DIR_ENTRIES_PER_SECTOR; i++) {
+			unsigned int entrypos = i * DIR_ENTRY_SIZE;
+
+			firstbyte = cached_buf[entrypos];
+			dir->entry++;
+
+			if (firstbyte == 0xe5) {
+				/*
+				 * free entry 
+				 */
+				sectoridx = 0;
+				dir->entrycount = 0;
+				continue;
+			}
+
+			if (firstbyte == 0) {
+				/*
+				 * last entry 
+				 */
+				entry->name[0] = 0;
+				dir->entrycount = 0;
+				return 0;
+			}
+
+			dir->entrycount++;
+
+			/*
+			 * longname entry? 
+			 */
+			if ((cached_buf[entrypos + FATDIR_ATTR] &
+			     FAT_ATTR_LONG_NAME_MASK) == FAT_ATTR_LONG_NAME) {
+				longarray[longs++] = entrypos + sectoridx;
+			} else {
+				if (parse_direntry
+				    (entry, &cached_buf[entrypos])) {
+
+					/*
+					 * don't return volume id entry 
+					 */
+					if (entry->attr == FAT_ATTR_VOLUME_ID)
+						continue;
+
+					/*
+					 * replace shortname with longname? 
+					 */
+					if (longs) {
+						int j, l = 0;
+						/*
+						 * iterate backwards through
+						 * the dir entries 
+						 */
+						for (j = longs - 1; j >= 0; j--) {
+							unsigned char *ptr =
+							    cached_buf;
+							int index =
+							    longarray[j];
+							/*
+							 * current or cached
+							 * sector? 
+							 */
+							if (sectoridx >=
+							    SECTOR_SIZE) {
+								if (sectoridx >=
+								    SECTOR_SIZE
+								    * 2) {
+									if ((index >= SECTOR_SIZE) && (index < SECTOR_SIZE * 2))
+										ptr = dir->sectorcache[1];
+									else
+										ptr = dir->sectorcache[2];
+								} else {
+									if (index < SECTOR_SIZE)
+										ptr = dir->sectorcache[1];
+								}
+
+								index &=
+								    SECTOR_SIZE
+								    - 1;
+							}
+
+							/*
+							 * names are stored in
+							 * unicode, but we only 
+							 * grab the low byte
+							 * (iso8859-1). 
+							 */
+							unicode2iso(ptr +
+								    index + 1,
+								    entry->
+								    name + l,
+								    5);
+							l += 5;
+							unicode2iso(ptr +
+								    index + 14,
+								    entry->
+								    name + l,
+								    6);
+							l += 6;
+							unicode2iso(ptr +
+								    index + 28,
+								    entry->
+								    name + l,
+								    2);
+							l += 2;
+						}
+						entry->name[l] = 0;
+					}
+					done = true;
+					sectoridx = 0;
+					i++;
+					break;
+				}
+			}
+		}
+
+		/*
+		 * save this sector, for longname use 
+		 */
+		if (sectoridx)
+			memcpy(dir->sectorcache[2], dir->sectorcache[0],
+			       SECTOR_SIZE);
+		else
+			memcpy(dir->sectorcache[1], dir->sectorcache[0],
+			       SECTOR_SIZE);
+		sectoridx += SECTOR_SIZE;
+
+	}
+	return 0;
+}
+#endif				/* #if defined(CFG_ROCKBOX_FAT) */
diff --git a/fs/fat/rockbox_file.c b/fs/fat/rockbox_file.c
new file mode 100644
index 0000000..1a2ee4f
--- /dev/null
+++ b/fs/fat/rockbox_file.c
@@ -0,0 +1,727 @@
+/*******************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: file.c,v 1.62 2005/01/28 21:32:16 hohensoh Exp $
+ *
+ * Copyright (C) 2002 by Bjrn Stenberg
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.66 of 'firmware/common/file.c' from rockbox CVS server.
+ *
+ ****************************************************************************/
+#include <common.h>
+#include <config.h>
+
+#if defined(CFG_ROCKBOX_FAT)
+#include <asm/errno.h>
+#include <asm/string.h>
+#include <fat.h>
+#include "rockbox_debug.h"
+
+#ifndef __HAVE_ARCH_STRNICMP
+extern int strnicmp(const char *s1, const char *s2, size_t len);
+#endif
+
+/*
+  These functions provide a roughly POSIX-compatible file IO API.
+
+  Since the fat32 driver only manages sectors, we maintain a one-sector
+  cache for each open file. This way we can provide byte access without
+  having to re-read the sector each time. 
+  The penalty is the RAM used for the cache and slightly more complex code.
+*/
+
+#if !defined(CFG_ROCKBOX_FAT_MAX_OPEN_FILES)
+#define CFG_ROCKBOX_FAT_MAX_OPEN_FILES 1
+#endif
+
+struct filedesc {
+	unsigned char cache[SECTOR_SIZE];
+	int cacheoffset;	/* invariant: 0 <= cacheoffset <= SECTOR_SIZE */
+	long fileoffset;
+	long size;
+	int attr;
+	struct fat_file fatfile;
+	bool busy;
+	bool write;
+	bool dirty;
+	bool trunc;
+};
+
+static struct filedesc openfiles[CFG_ROCKBOX_FAT_MAX_OPEN_FILES];
+
+static int flush_sector_cache(int fd);
+
+int
+creat(const char *pathname, mode_t mode)
+{
+	(void) mode;
+	return open(pathname, O_WRONLY | O_CREAT | O_TRUNC);
+}
+
+int
+open(const char *pathname, int flags)
+{
+	DIR *dir;
+	struct dirent *entry;
+	int fd;
+	char pathnamecopy[MAX_PATH];
+	char *name;
+	struct filedesc *file = NULL;
+	int rc;
+
+	fat_dprintf("pathname \"%s\", flags %d\n", pathname, flags);
+
+	/*
+	 * Only absolute paths are supported.
+	 */
+	if (pathname[0] != '/') {
+		fat_deprintf("\"%s\" is not an absolute path\n", pathname);
+		errno = EINVAL;
+		return -1;
+	}
+
+	/*
+	 * find a free file descriptor 
+	 */
+	for (fd = 0; fd < CFG_ROCKBOX_FAT_MAX_OPEN_FILES; fd++)
+		if (!openfiles[fd].busy)
+			break;
+
+	if (fd == CFG_ROCKBOX_FAT_MAX_OPEN_FILES) {
+		fat_deprintf("Too many files open\n");
+		errno = EMFILE;
+		return -2;
+	}
+
+	file = &openfiles[fd];
+	memset(file, 0, sizeof (struct filedesc));
+
+	if (flags & (O_RDWR | O_WRONLY)) {
+		file->write = true;
+
+		if (flags & O_TRUNC)
+			file->trunc = true;
+	}
+	file->busy = true;
+
+	strncpy(pathnamecopy, pathname, sizeof (pathnamecopy));
+	pathnamecopy[sizeof (pathnamecopy) - 1] = 0;
+
+	/*
+	 * locate filename 
+	 */
+	name = strrchr(pathnamecopy + 1, '/');
+	if (name) {
+		*name = 0;
+		dir = opendir(pathnamecopy);
+		*name = '/';
+		name++;
+	} else {
+		dir = opendir("/");
+		name = pathnamecopy + 1;
+	}
+	if (!dir) {
+		fat_deprintf("Failed to open directory\n");
+		errno = EIO;
+		file->busy = false;
+		return -4;
+	}
+
+	if (name[0] == 0) {
+		fat_deprintf("Empty file name\n");
+		errno = EINVAL;
+		file->busy = false;
+		closedir(dir);
+		return -5;
+	}
+
+	/*
+	 * scan dir for name 
+	 */
+	while ((entry = readdir(dir))) {
+		if (!strnicmp(name, entry->d_name, strlen(name))) {
+			fat_open(entry->startcluster,
+				 &(file->fatfile), &(dir->fatdir));
+			file->size = file->trunc ? 0 : entry->size;
+			file->attr = entry->attribute;
+			break;
+		}
+	}
+
+	if (!entry) {
+		fat_dprintf("Didn't find file \"%s\"\n", name);
+		if (file->write && (flags & O_CREAT)) {
+			rc = fat_create_file(name, &(file->fatfile),
+					     &(dir->fatdir));
+			if (rc < 0) {
+				fat_deprintf
+				    ("Couldn't create \"%s\" in \"%s\"\n", name,
+				     pathnamecopy);
+				errno = EIO;
+				file->busy = false;
+				closedir(dir);
+				return rc * 10 - 6;
+			}
+			file->size = 0;
+			file->attr = 0;
+		} else {
+			fat_deprintf("Couldn't find \"%s\" in \"%s\"\n",
+				     name, pathnamecopy);
+			errno = ENOENT;
+			file->busy = false;
+			closedir(dir);
+			return -7;
+		}
+	} else {
+		if (file->write && (file->attr & FAT_ATTR_DIRECTORY)) {
+			errno = EISDIR;
+			file->busy = false;
+			closedir(dir);
+			return -8;
+		}
+	}
+	closedir(dir);
+
+	file->cacheoffset = -1;
+	file->fileoffset = 0;
+
+	if (file->write && (flags & O_APPEND)) {
+		rc = lseek(fd, 0, SEEK_END);
+		if (rc < 0)
+			return rc * 10 - 9;
+	}
+
+	return fd;
+}
+
+int
+close(int fd)
+{
+	struct filedesc *file = &openfiles[fd];
+	int rc = 0;
+
+	fat_dprintf("fd %d\n", fd);
+
+	if (fd < 0 || fd > CFG_ROCKBOX_FAT_MAX_OPEN_FILES - 1) {
+		errno = EINVAL;
+		return -1;
+	}
+	if (!file->busy) {
+		errno = EBADF;
+		return -2;
+	}
+	if (file->write) {
+		rc = fsync(fd);
+		if (rc < 0)
+			return rc * 10 - 3;
+	}
+
+	file->busy = false;
+	return 0;
+}
+
+int
+fsync(int fd)
+{
+	struct filedesc *file = &openfiles[fd];
+	int rc = 0;
+
+	fat_dprintf("fd %d\n", fd);
+
+	if (fd < 0 || fd > CFG_ROCKBOX_FAT_MAX_OPEN_FILES - 1) {
+		errno = EINVAL;
+		return -1;
+	}
+	if (!file->busy) {
+		errno = EBADF;
+		return -2;
+	}
+	if (file->write) {
+		/*
+		 * flush sector cache 
+		 */
+		if (file->dirty) {
+			rc = flush_sector_cache(fd);
+			if (rc < 0)
+				return rc * 10 - 3;
+		}
+
+		/*
+		 * truncate? 
+		 */
+		if (file->trunc) {
+			rc = ftruncate(fd, file->size);
+			if (rc < 0)
+				return rc * 10 - 4;
+		}
+
+		/*
+		 * tie up all loose ends 
+		 */
+		rc = fat_closewrite(&(file->fatfile), file->size, file->attr);
+		if (rc < 0)
+			return rc * 10 - 5;
+	}
+	return 0;
+}
+
+int
+remove(const char *name)
+{
+	int rc;
+	struct filedesc *file;
+	int fd = open(name, O_WRONLY);
+	if (fd < 0)
+		return fd * 10 - 1;
+
+	file = &openfiles[fd];
+	rc = fat_remove(&(file->fatfile));
+	if (rc < 0) {
+		fat_deprintf("Failed to remove file (rc = %d)\n", rc);
+		errno = EIO;
+		return rc * 10 - 3;
+	}
+
+	file->size = 0;
+
+	rc = close(fd);
+	if (rc < 0)
+		return rc * 10 - 4;
+
+	return 0;
+}
+
+int
+rename(const char *path, const char *newpath)
+{
+	int rc, fd;
+	DIR *dir;
+	char *nameptr;
+	char *dirptr;
+	struct filedesc *file;
+	char newpath2[MAX_PATH];
+
+	/*
+	 * verify new path does not already exist 
+	 */
+	/*
+	 * If it is a directory, errno == EISDIR if the name exists 
+	 */
+	fd = open(newpath, O_RDONLY);
+	if (fd >= 0 || errno == EISDIR) {
+		close(fd);
+		errno = EBUSY;
+		return -1;
+	}
+	close(fd);
+
+	fd = open(path, O_RDONLY);
+	if (fd < 0) {
+		errno = EIO;
+		return fd * 10 - 2;
+	}
+
+	/*
+	 * extract new file name 
+	 */
+	nameptr = strrchr(newpath, '/');
+	if (nameptr)
+		nameptr++;
+	else
+		return -3;
+
+	/*
+	 * Extract new path 
+	 */
+	strcpy(newpath2, newpath);
+
+	dirptr = strrchr(newpath2, '/');
+	if (dirptr)
+		*dirptr = 0;
+	else
+		return -4;
+
+	dirptr = newpath2;
+
+	if (strlen(dirptr) == 0) {
+		dirptr = "/";
+	}
+
+	dir = opendir(dirptr);
+	if (!dir)
+		return -5;
+
+	file = &openfiles[fd];
+	rc = fat_rename(&file->fatfile, &dir->fatdir, nameptr,
+			file->size, file->attr);
+	if (rc < 0) {
+		fat_deprintf("Failed to remane file\n", rc);
+		errno = EIO;
+		return rc * 10 - 6;
+	}
+
+	rc = close(fd);
+	if (rc < 0) {
+		errno = EIO;
+		return rc * 10 - 7;
+	}
+
+	rc = closedir(dir);
+	if (rc < 0) {
+		errno = EIO;
+		return rc * 10 - 8;
+	}
+
+	return 0;
+}
+
+int
+ftruncate(int fd, off_t size)
+{
+	int rc, sector;
+	struct filedesc *file = &openfiles[fd];
+
+	sector = size / SECTOR_SIZE;
+	if (size % SECTOR_SIZE)
+		sector++;
+
+	rc = fat_seek(&(file->fatfile), sector);
+	if (rc < 0) {
+		errno = EIO;
+		return rc * 10 - 1;
+	}
+
+	rc = fat_truncate(&(file->fatfile));
+	if (rc < 0) {
+		errno = EIO;
+		return rc * 10 - 2;
+	}
+
+	file->size = size;
+
+	return 0;
+}
+
+static int
+flush_sector_cache(int fd)
+{
+	int rc;
+	struct filedesc *file = &openfiles[fd];
+	long sector = file->fileoffset / SECTOR_SIZE;
+
+	fat_dprintf("Flush dirty sectors\n");
+
+	/*
+	 * make sure we are on correct sector 
+	 */
+	rc = fat_seek(&(file->fatfile), sector);
+	if (rc < 0)
+		return rc * 10 - 3;
+
+	rc = fat_readwrite(&(file->fatfile), 1, file->cache, true);
+
+	if (rc < 0) {
+		if (file->fatfile.eof)
+			errno = ENOSPC;
+
+		return rc * 10 - 2;
+	}
+
+	file->dirty = false;
+
+	return 0;
+}
+
+static int
+readwrite(int fd, void *buf, long count, bool write)
+{
+	long sectors;
+	long nread = 0;
+	struct filedesc *file = &openfiles[fd];
+	int rc;
+
+	if (!file->busy) {
+		errno = EBADF;
+		return -1;
+	}
+
+	fat_dprintf("fd %d, buf %lx, count %ld, %s\n",
+		    fd, (long) buf, count, write ? "write" : "read");
+
+	/*
+	 * attempt to read past EOF? 
+	 */
+	if (!write && count > file->size - file->fileoffset)
+		count = file->size - file->fileoffset;
+
+	/*
+	 * any head bytes? 
+	 */
+	if (file->cacheoffset != -1) {
+		int headbytes;
+		int offs = file->cacheoffset;
+		if (count <= SECTOR_SIZE - file->cacheoffset) {
+			headbytes = count;
+			file->cacheoffset += count;
+			if (file->cacheoffset >= SECTOR_SIZE)
+				file->cacheoffset = -1;
+		} else {
+			headbytes = SECTOR_SIZE - file->cacheoffset;
+			file->cacheoffset = -1;
+		}
+
+		if (write) {
+			memcpy(file->cache + offs, buf, headbytes);
+			if (offs + headbytes == SECTOR_SIZE) {
+				int rc = flush_sector_cache(fd);
+				if (rc < 0) {
+					errno = EIO;
+					return rc * 10 - 2;
+				}
+				file->cacheoffset = -1;
+			} else
+				file->dirty = true;
+		} else {
+			memcpy(buf, file->cache + offs, headbytes);
+		}
+
+		nread = headbytes;
+		count -= headbytes;
+	}
+
+	/*
+	 * If the buffer has been modified, either it has been flushed already
+	 * (if (offs+headbytes == SECTOR_SIZE)...) or does not need to be (no
+	 * more data to follow in this call). Do NOT flush here. 
+	 */
+
+	/*
+	 * read/write whole sectors right into/from the supplied buffer 
+	 */
+	sectors = count / SECTOR_SIZE;
+	if (sectors) {
+		int rc = fat_readwrite(&(file->fatfile), sectors,
+				       (unsigned char *) buf + nread, write);
+		if (rc < 0) {
+			fat_deprintf("Failed to %s %ld sectors\n",
+				     write ? "write" : "read", sectors);
+			errno = EIO;
+			if (write && file->fatfile.eof) {
+				fat_deprintf("No space left on device\n");
+				errno = ENOSPC;
+			} else {
+				file->fileoffset += nread;
+			}
+			file->cacheoffset = -1;
+			return nread ? nread : rc * 10 - 4;
+		} else {
+			if (rc > 0) {
+				nread += rc * SECTOR_SIZE;
+				count -= sectors * SECTOR_SIZE;
+
+				/*
+				 * if eof, skip tail bytes 
+				 */
+				if (rc < sectors)
+					count = 0;
+			} else {
+				/*
+				 * eof 
+				 */
+				count = 0;
+			}
+
+			file->cacheoffset = -1;
+		}
+	}
+
+	/*
+	 * any tail bytes? 
+	 */
+	if (count) {
+		if (write) {
+			if (file->fileoffset + nread < file->size) {
+				/*
+				 * sector is only partially filled. copy-back
+				 * from disk 
+				 */
+				int rc;
+				fat_dprintf("Copy-back tail cache\n");
+				rc = fat_readwrite(&(file->fatfile), 1,
+						   file->cache, false);
+				if (rc < 0) {
+					fat_deprintf
+					    ("Failed to write cache to disk\n");
+					errno = EIO;
+					file->fileoffset += nread;
+					file->cacheoffset = -1;
+					return nread ? nread : rc * 10 - 5;
+				}
+				/*
+				 * seek back one sector to put file position
+				 * right 
+				 */
+				rc = fat_seek(&(file->fatfile),
+					      (file->fileoffset + nread) /
+					      SECTOR_SIZE);
+				if (rc < 0) {
+					fat_deprintf("fat_seek() failed\n", rc);
+					errno = EIO;
+					file->fileoffset += nread;
+					file->cacheoffset = -1;
+					return nread ? nread : rc * 10 - 6;
+				}
+			}
+			memcpy(file->cache, (unsigned char *) buf + nread,
+			       count);
+			file->dirty = true;
+		} else {
+			rc = fat_readwrite(&(file->fatfile), 1, &(file->cache),
+					   false);
+			if (rc < 1) {
+				fat_deprintf("Failed to cache sector\n", rc);
+				errno = EIO;
+				file->fileoffset += nread;
+				file->cacheoffset = -1;
+				return nread ? nread : rc * 10 - 7;
+			}
+			memcpy((unsigned char *) buf + nread, file->cache,
+			       count);
+		}
+
+		nread += count;
+		file->cacheoffset = count;
+	}
+
+	file->fileoffset += nread;
+	fat_dprintf("fileoffset %ld\n", file->fileoffset);
+
+	/*
+	 * adjust file size to length written 
+	 */
+	if (write && file->fileoffset > file->size)
+		file->size = file->fileoffset;
+
+	return nread;
+}
+
+ssize_t
+write(int fd, const void *buf, size_t count)
+{
+	if (!openfiles[fd].write) {
+		errno = EACCES;
+		return -1;
+	}
+	return readwrite(fd, (void *) buf, count, true);
+}
+
+ssize_t
+read(int fd, void *buf, size_t count)
+{
+	return readwrite(fd, buf, count, false);
+}
+
+off_t
+lseek(int fd, off_t offset, int whence)
+{
+	off_t pos;
+	long newsector;
+	long oldsector;
+	int sectoroffset;
+	int rc;
+	struct filedesc *file = &openfiles[fd];
+
+	fat_dprintf("fd %d, offset %ld, whence %d\n", fd, offset, whence);
+
+	if (!file->busy) {
+		errno = EBADF;
+		return -1;
+	}
+
+	switch (whence) {
+	case SEEK_SET:
+		pos = offset;
+		break;
+
+	case SEEK_CUR:
+		pos = file->fileoffset + offset;
+		break;
+
+	case SEEK_END:
+		pos = file->size + offset;
+		break;
+
+	default:
+		errno = EINVAL;
+		return -2;
+	}
+	if ((pos < 0) || (pos > file->size)) {
+		errno = EINVAL;
+		return -3;
+	}
+
+	/*
+	 * new sector? 
+	 */
+	newsector = pos / SECTOR_SIZE;
+	oldsector = file->fileoffset / SECTOR_SIZE;
+	sectoroffset = pos % SECTOR_SIZE;
+
+	if ((newsector != oldsector) ||
+	    ((file->cacheoffset == -1) && sectoroffset)) {
+
+		if (newsector != oldsector) {
+			if (file->dirty) {
+				rc = flush_sector_cache(fd);
+				if (rc < 0)
+					return rc * 10 - 5;
+			}
+
+			rc = fat_seek(&(file->fatfile), newsector);
+			if (rc < 0) {
+				errno = EIO;
+				return rc * 10 - 4;
+			}
+		}
+		if (sectoroffset) {
+			rc = fat_readwrite(&(file->fatfile), 1,
+					   &(file->cache), false);
+			if (rc < 0) {
+				errno = EIO;
+				return rc * 10 - 6;
+			}
+			file->cacheoffset = sectoroffset;
+		} else
+			file->cacheoffset = -1;
+	} else if (file->cacheoffset != -1)
+		file->cacheoffset = sectoroffset;
+
+	file->fileoffset = pos;
+
+	return pos;
+}
+
+off_t
+filesize(int fd)
+{
+	struct filedesc *file = &openfiles[fd];
+
+	if (!file->busy) {
+		errno = EBADF;
+		return -1;
+	}
+
+	return file->size;
+}
+#endif				/* #if defined(CFG_ROCKBOX_FAT) */
diff --git a/fs/fat/rockbox_wrapper.c b/fs/fat/rockbox_wrapper.c
new file mode 100644
index 0000000..fa06018
--- /dev/null
+++ b/fs/fat/rockbox_wrapper.c
@@ -0,0 +1,508 @@
+/*------------------------------------------------------------------------------
+ * (C) Copyright 2006
+ * Keith Outwater, (outwater4@comcast.net)
+ *
+ * See file CREDITS for list of people who contributed to this
+ * project.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+ * MA 02111-1307 USA
+ */
+
+#include <common.h>
+#include <config.h>
+
+#if defined(CFG_ROCKBOX_FAT)
+
+/*
+ * FIXME: The pwd, cd and file path builder code needs to be fixed/implemented.
+ */
+
+#include <asm/errno.h>
+#include <asm/string.h>
+#include <fat.h>
+#include <linux/string.h>
+#include "rockbox_debug.h"
+
+const char *month[] = {
+	"(null)",
+	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
+	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
+};
+
+/*
+ * The current working directory.
+ * Should we keep track of this on a per-partiton basis?
+ */
+char file_cwd[ROCKBOX_CWD_LEN + 1] = "/";
+
+#ifndef __HAVE_ARCH_STRNICMP
+/*
+ * Convert a character to lower case
+ */
+__inline__ static char
+_tolower(char c)
+{
+	if ((c >= 'A') && (c <= 'Z')) {
+		c = (c - 'A') + 'a';
+	}
+	return c;
+}
+
+/**
+ * strnicmp - Case insensitive, length-limited string comparison
+ * @s1: One string
+ * @s2: The other string
+ * @len: the maximum number of characters to compare
+ */
+int
+strnicmp(const char *s1, const char *s2, size_t len)
+{
+	/*
+	 * Yes, Virginia, it had better be unsigned 
+	 */
+	unsigned char c1, c2;
+
+	c1 = 0;
+	c2 = 0;
+	if (len) {
+		do {
+			c1 = *s1;
+			c2 = *s2;
+			s1++;
+			s2++;
+			if (!c1)
+				break;
+			if (!c2)
+				break;
+			if (c1 == c2)
+				continue;
+			c1 = _tolower(c1);
+			c2 = _tolower(c2);
+			if (c1 != c2)
+				break;
+		} while (--len);
+	}
+	return (int) c1 - (int) c2;
+}
+#endif
+
+/*
+ * U-Boot 'fat ls' wrapper. Code lifted from rockbox "filetree.c"
+ * Return number of bytes read on success, -1 on error.
+ */
+int
+file_fat_ls(const char *dirname)
+{
+	int i;
+	int name_buffer_used = 0;
+	DIR *dir;
+
+	struct tree_context {
+		int filesindir;
+		int dirsindir;
+		int dirlength;
+		void *dircache;
+		char name_buffer[2048];
+		int name_buffer_size;
+		bool dirfull;
+	} c;
+
+	struct entry {
+		short attr;	/* FAT attributes + file type flags */
+		unsigned long time_write;	/* Last write time */
+		char *name;
+	};
+
+	if (strlen(dirname) > ROCKBOX_CWD_LEN) {
+		errno = ENAMETOOLONG;
+		return -1;
+	}
+
+	dir = opendir(dirname);
+	if (!dir) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Error: ls() of \"%s\" failed: %s\n",
+			dirname, strerror(errno));
+#endif
+		return -1;
+	}
+
+	c.dirsindir = 0;
+	c.dirfull = false;
+	c.name_buffer_size = sizeof (c.name_buffer) - 1;
+
+	/*
+	 * FIXME: figure out max i 
+	 */
+	for (i = 0; i < 10; i++) {
+		int len;
+		struct dirent *entry = readdir(dir);
+		struct entry *dptr =
+		    (struct entry *) (c.dircache + i * sizeof (struct entry));
+
+		if (!entry)
+			break;
+
+		len = strlen(entry->d_name);
+		printf("%9lu %s %02d %4d %02d:%02d:%02d %s\n",
+		       entry->size,
+		       month[(entry->wrtdate >> 5) & 0xf],
+		       entry->wrtdate & 0xf,
+		       ((entry->wrtdate >> 9) & 0x7f) + 1980,
+		       (entry->wrttime >> 11) & 0x1f,
+		       (entry->wrttime >> 5) & 0x3f, entry->wrttime & 0x1f,
+		       entry->d_name);
+#if 0
+		/*
+		 * skip directories . and ..
+		 */
+		if ((entry->attribute & FAT_ATTR_DIRECTORY) &&
+		    (((len == 1) &&
+		      (!strncmp(entry->d_name, ".", 1))) ||
+		     ((len == 2) && (!strncmp(entry->d_name, "..", 2))))) {
+			i--;
+			continue;
+		}
+
+		/*
+		 * Skip FAT volume ID
+		 */
+		if (entry->attribute & FAT_ATTR_VOLUME_ID) {
+			i--;
+			continue;
+		}
+#endif
+
+		dptr->attr = entry->attribute;
+		if (len > c.name_buffer_size - name_buffer_used - 1) {
+			/*
+			 * Tell the world that we ran out of buffer space 
+			 */
+			c.dirfull = true;
+			fat_dprintf("Dir buffer is full\n");
+			break;
+		}
+
+		dptr->name = &c.name_buffer[name_buffer_used];
+		dptr->time_write =
+		    (long) entry->wrtdate << 16 | (long) entry->wrttime;
+		strcpy(dptr->name, entry->d_name);
+		name_buffer_used += len + 1;
+
+		/*
+		 * count the remaining dirs 
+		 */
+		if (dptr->attr & FAT_ATTR_DIRECTORY)
+			c.dirsindir++;
+	}
+
+	c.filesindir = i;
+	c.dirlength = i;
+	closedir(dir);
+	printf("\n%d file%s %d dir%s",
+	       c.filesindir, c.filesindir == 1 ? "," : "s,",
+	       c.dirsindir, c.dirsindir == 1 ? "\n" : "s\n");
+	return 0;
+}
+
+/*
+ * U-Boot 'file read' wrapper.
+ * Return number of bytes read on success, -1 on error.
+ */
+long
+file_fat_read(const char *filename, void *buffer, unsigned long maxsize)
+{
+	int fd;
+	long bytes;
+	long file_size;
+
+	printf("Reading %s\n", filename);
+	fd = open(filename, O_RDONLY);
+	if (fd < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Error: open() of \"%s\" failed: %s\n",
+			filename, strerror(errno));
+#endif
+		return -1;
+	}
+
+	/*
+	 * If the number of bytes to read is zero, read the entire file.
+	 */
+	if (maxsize == 0) {
+		file_size = filesize(fd);
+		if (file_size < 0) {
+			fat_eprintf("Call to filesize() failed\n");
+			return -1;
+		}
+
+		maxsize = (unsigned long) file_size;
+		fat_dprintf("Reading entire file (%lu bytes)\n", maxsize);
+	}
+
+	bytes = (long) read(fd, buffer, (size_t) maxsize);
+	if ((unsigned long) bytes != maxsize) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Read error: %s\n", strerror(errno));
+#endif
+		close(fd);
+		return -1;
+	}
+
+	close(fd);
+	return bytes;
+}
+
+/*
+ * U-Boot 'file write' wrapper.
+ * Return number of bytes written on success, -1 on error.
+ */
+long
+file_fat_write(const char *filename, void *buffer, unsigned long maxsize)
+{
+	int fd;
+	long bytes;
+
+	printf("Writing %s\n", filename);
+	fd = open(filename, O_WRONLY | O_CREAT);
+	if (fd < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Open of \"%s\" failed: %s\n",
+			filename, strerror(errno));
+#endif
+		return -1;
+	}
+
+	bytes = (long) write(fd, buffer, (size_t) maxsize);
+	if ((unsigned long) bytes != maxsize) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Write failed: %s\n", strerror(errno));
+#endif
+		close(fd);
+		return -1;
+	}
+
+	close(fd);
+	return bytes;
+}
+
+int
+file_fat_pwd(void)
+{
+	printf("%s\n", file_cwd);
+	return 1;
+}
+
+/*
+ * U-Boot 'file rm' wrapper.
+ * Return 1 on success, -1 on error and set 'errno'.
+ */
+int
+file_fat_rm(const char *filename)
+{
+	int rc;
+
+	fat_dprintf("Remove \"%s\"\n", filename);
+	rc = remove(filename);
+	if (rc < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Removal of \"%s\" failed: %s\n",
+			filename, strerror(errno));
+#endif
+		return -1;
+	}
+
+	return 1;
+}
+
+/*
+ * U-Boot 'file mkdir' wrapper.
+ * Return 1 on success, -1 on error and set 'errno'.
+ */
+int
+file_fat_mkdir(const char *dirname)
+{
+	int rc;
+
+	fat_dprintf("Make dir \"%s\"\n", dirname);
+	rc = mkdir(dirname, 0);
+	if (rc < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Creation of \"%s\" failed: %s\n",
+			dirname, strerror(errno));
+#endif
+		return -1;
+	}
+
+	return 1;
+}
+
+/*
+ * U-Boot 'file rmdir' wrapper.
+ * Return 1 on success, -1 on error and set 'errno'.
+ */
+int
+file_fat_rmdir(const char *dirname)
+{
+	int rc;
+
+	fat_dprintf("Remove dir \"%s\"\n", dirname);
+	rc = rmdir(dirname);
+	if (rc < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Removal of \"%s\" failed: %s\n",
+			dirname, strerror(errno));
+#endif
+		return -1;
+	}
+
+	return 1;
+}
+
+static void
+pathcpy(char *dest, const char *src)
+{
+	char *origdest = dest;
+
+	do {
+		if (dest - file_cwd >= CWD_LEN) {
+			*dest = '\0';
+			return;
+		}
+
+		*(dest) = *(src);
+		if (*src == '\0') {
+			if (dest-- != origdest && ISDIRDELIM(*dest)) {
+				*dest = '\0';
+			}
+			return;
+		}
+
+		++dest;
+		if (ISDIRDELIM(*src)) {
+			while (ISDIRDELIM(*src))
+				src++;
+		} else {
+			src++;
+		}
+	} while (1);
+}
+
+/*
+ * U-Boot 'file cd' wrapper.
+ */
+int
+file_fat_cd(const char *path)
+{
+	if (ISDIRDELIM(*path)) {
+		while (ISDIRDELIM(*path))
+			path++;
+		strncpy(file_cwd + 1, path, CWD_LEN - 1);
+	} else {
+		const char *origpath = path;
+		char *tmpstr = file_cwd;
+		int back = 0;
+
+		while (*tmpstr != '\0')
+			tmpstr++;
+
+		do {
+			tmpstr--;
+		} while (ISDIRDELIM(*tmpstr));
+
+		while (*path == '.') {
+			path++;
+			while (*path == '.') {
+				path++;
+				back++;
+			}
+
+			if (*path != '\0' && !ISDIRDELIM(*path)) {
+				path = origpath;
+				back = 0;
+				break;
+			}
+
+			while (ISDIRDELIM(*path))
+				path++;
+
+			origpath = path;
+		}
+
+		while (back--) {
+			/*
+			 * Strip off path component 
+			 */
+			while (!ISDIRDELIM(*tmpstr)) {
+				tmpstr--;
+			}
+
+			if (tmpstr == file_cwd) {
+				/*
+				 * Incremented again right after the loop. 
+				 */
+				tmpstr--;
+				break;
+			}
+
+			/*
+			 * Skip delimiters 
+			 */
+			while (ISDIRDELIM(*tmpstr))
+				tmpstr--;
+		}
+
+		tmpstr++;
+		if (*path == '\0') {
+			if (tmpstr == file_cwd) {
+				*tmpstr = '/';
+				tmpstr++;
+			}
+
+			*tmpstr = '\0';
+			return 0;
+		}
+
+		*tmpstr = '/';
+		pathcpy(tmpstr + 1, path);
+	}
+
+	printf("New CWD is '%s'\n", file_cwd);
+	return 0;
+}
+
+/*
+ * U-Boot 'file mv' wrapper.
+ * Return 1 on success, -1 on error and set 'errno'.
+ */
+int
+file_fat_mv(const char *oldname, const char *newpath)
+{
+	int rc;
+
+	fat_dprintf("Move \'%s\" to \"%s\"\n", oldname, newpath);
+	rc = rename(oldname, newpath);
+	if (rc < 0) {
+#ifdef CONFIG_STRERROR
+		fprintf(stderr, "Failed to move \"%s\" to \"%s\" : %s\n",
+			oldname, newpath, strerror(errno));
+#endif
+		return -1;
+	}
+
+	return 1;
+}
+#endif				/* #if defined(CFG_ROCKBOX_FAT) */
diff --git a/include/fat.h b/include/fat.h
index 0645458..734d217 100644
--- a/include/fat.h
+++ b/include/fat.h
@@ -3,6 +3,8 @@
  *
  * 2002-07-28 - rjones@nexus-tech.net - ported to ppcboot v1.1.6
  * 2003-03-10 - kharris@nexus-tech.net - ported to u-boot
+ * 2006-01-18 - Keith Outwater (outwater@comcast.net) - added write support
+ *    using FAT driver from rockbox project (www.rockbox.org)
  *
  * See file CREDITS for list of people who contributed to this
  * project.
@@ -80,7 +82,7 @@ #define LS_YES	1
 #define LS_DIR	1
 #define LS_ROOT	2
 
-#ifdef DEBUG
+#if (defined(DEBUG) || defined(FAT_DEBUG))
 #define FAT_DPRINT(args...)	printf(args)
 #else
 #define FAT_DPRINT(args...)
@@ -113,6 +115,14 @@ #define START(dent)	(FAT2CPU16((dent)->s
 			  (FAT2CPU16((dent)->starthi) << 16)))
 
 
+#ifdef CFG_ROCKBOX_FAT
+/*
+ * Remember the current working directory for the four primary partitions.
+ */
+#define	NUM_PARTS_WITH_CWD	4
+#define CWD_LEN			511
+#endif
+
 typedef struct boot_sector {
 	__u8	ignored[3];	/* Bootstrap code */
 	char	system_id[8];	/* Name of fs */
@@ -187,11 +197,20 @@ typedef struct {
 	int	fatbufnum;	/* Used by get_fatent, init to -1 */
 } fsdata;
 
+/*
+ * This struct encapsulates partiton and device info for the currently active
+ * block device.
+ */
+typedef struct cur_block_dev {
+	block_dev_desc_t *cur_dev;
+	unsigned long part_offset;
+	int cur_part;
+} cur_block_dev_t;
+
 typedef int	(file_detectfs_func)(void);
 typedef int	(file_ls_func)(const char *dir);
 typedef long	(file_read_func)(const char *filename, void *buffer,
 				 unsigned long maxsize);
-
 struct filesystem {
 	file_detectfs_func *detect;
 	file_ls_func	   *ls;
@@ -204,12 +223,28 @@ file_detectfs_func	file_fat_detectfs;
 file_ls_func		file_fat_ls;
 file_read_func		file_fat_read;
 
-/* Currently this doesn't check if the dir exists or is valid... */
+/*Currently this doesn't check if the dir exists or is valid */
 int file_cd(const char *path);
+
 int file_fat_detectfs(void);
 int file_fat_ls(const char *dir);
 long file_fat_read(const char *filename, void *buffer, unsigned long maxsize);
+#ifdef CFG_ROCKBOX_FAT
+long file_fat_write(const char *filename, void *buffer, unsigned long filesize);
+int file_fat_rm(const char *filename);
+int file_fat_mkdir(const char *dirname);
+int file_fat_rmdir(const char *dirname);
+int file_fat_pwd(void);
+int file_fat_cd(const char *dirname);
+int file_fat_mv(const char *oldname, const char *newname);
+#endif
 const char *file_getfsname(int idx);
 int fat_register_device(block_dev_desc_t *dev_desc, int part_no);
 
+#ifdef CFG_ROCKBOX_FAT
+#ifndef ROCKBOX_FAT_H
+#include <rockbox_fat.h>
+#endif
+#endif
+
 #endif /* _FAT_H_ */
diff --git a/include/part.h b/include/part.h
index 318aa3c..9b2cd3e 100644
--- a/include/part.h
+++ b/include/part.h
@@ -35,7 +35,7 @@ typedef struct block_dev_desc {
 #ifdef CONFIG_LBA48
 	unsigned char	lba48;		/* device can use 48bit addr (ATA/ATAPI v7) */
 #endif
-	lbaint_t		lba;	  	/* number of blocks */
+	lbaint_t	lba;	  	/* number of blocks */
 	unsigned long	blksz;		/* block size */
 	unsigned char	vendor [40+1]; 	/* IDE model, SCSI Vendor */
 	unsigned char	product[20+1];	/* IDE Serial no, SCSI product */
@@ -44,6 +44,12 @@ #endif
 				      unsigned long start,
 				      lbaint_t blkcnt,
 				      unsigned long *buffer);
+#ifdef CFG_ROCKBOX_FAT
+	unsigned long	(*block_write)(int dev,
+				       unsigned long start,
+				       lbaint_t blkcnt,
+				       unsigned long *buffer);
+#endif
 }block_dev_desc_t;
 
 /* Interface types: */
@@ -54,6 +60,7 @@ #define IF_TYPE_ATAPI		3
 #define IF_TYPE_USB		4
 #define IF_TYPE_DOC		5
 #define IF_TYPE_MMC		6
+#define IF_TYPE_SYSACE		7
 
 /* Part types */
 #define PART_TYPE_UNKNOWN	0x00
diff --git a/include/rockbox_dir.h b/include/rockbox_dir.h
new file mode 100644
index 0000000..56965f1
--- /dev/null
+++ b/include/rockbox_dir.h
@@ -0,0 +1,62 @@
+/***************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: dir.h,v 1.12 2005/02/26 21:18:05 jyp Exp $
+ *
+ * Copyright (C) 2002 by Bjrn Stenberg
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.12 of 'firmware/common/dir.h' from rockbox CVS server.
+ *
+ ****************************************************************************/
+
+#ifndef _ROCKBOX_DIR_H_
+#define _ROCKBOX_DIR_H_
+
+#include <stdbool.h>
+#include <rockbox_file.h>
+
+#ifndef DIRENT_DEFINED
+
+struct dirent {
+	unsigned char d_name[MAX_PATH];
+	int attribute;
+	long size;
+	long startcluster;
+	unsigned short wrtdate;	/* Last write date */
+	unsigned short wrttime;	/* Last write time */
+};
+#endif
+
+#include <rockbox_fat.h>
+
+typedef struct {
+	bool busy;
+	long startcluster;
+	struct fat_dir fatdir;
+	struct fat_dir parent_dir;
+	struct dirent theent;
+} DIR;
+
+#ifndef DIRFUNCTIONS_DEFINED
+
+extern DIR *opendir(const char *name);
+extern int closedir(DIR * dir);
+extern int mkdir(const char *name, int mode);
+extern int rmdir(const char *name);
+
+extern struct dirent *readdir(DIR * dir);
+
+#endif				/* DIRFUNCTIONS_DEFINED */
+
+#endif				/* #ifndef _ROCKBOX_DIR_H_ */
diff --git a/include/rockbox_fat.h b/include/rockbox_fat.h
new file mode 100644
index 0000000..e9f705a
--- /dev/null
+++ b/include/rockbox_fat.h
@@ -0,0 +1,109 @@
+/***************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: fat.h,v 1.13 2005/02/26 21:18:05 jyp Exp $
+ *
+ * Copyright (C) 2002 by Linus Nielsen Feltzing
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.13 of 'firmware/export/fat.h' from rockbox CVS server.
+ *
+ ****************************************************************************/
+
+#ifndef ROCKBOX_FAT_H
+#define ROCKBOX_FAT_H
+
+#include <stdbool.h>
+
+/*
+ * The max length of the current working dorectory.
+ */
+#define	ROCKBOX_CWD_LEN	512
+
+struct fat_direntry {
+	unsigned char name[256];	/* Name plus \0 */
+	unsigned short attr;		/* Attributes */
+	unsigned char crttimetenth;	/* Millisecond creation time stamp
+					 * (0-199) */
+	unsigned short crttime;		/* Creation time */
+	unsigned short crtdate;		/* Creation date */
+	unsigned short lstaccdate;	/* Last access date */
+	unsigned short wrttime;		/* Last write time */
+	unsigned short wrtdate;		/* Last write date */
+	unsigned long filesize;		/* File size in bytes */
+	long firstcluster;		/* fstclusterhi<<16 + fstcluslo */
+};
+
+#define FAT_ATTR_READ_ONLY   0x01
+#define FAT_ATTR_HIDDEN      0x02
+#define FAT_ATTR_SYSTEM      0x04
+#define FAT_ATTR_VOLUME_ID   0x08
+#define FAT_ATTR_DIRECTORY   0x10
+#define FAT_ATTR_ARCHIVE     0x20
+#define FAT_ATTR_VOLUME      0x40	/* this is a volume, not a real dir */
+
+struct fat_file {
+	long firstcluster;	/* first cluster in file */
+	long lastcluster;	/* cluster of last access */
+	long lastsector;	/* sector of last access */
+	long clusternum;	/* current clusternum */
+	long sectornum;		/* sector number in this cluster */
+	unsigned int direntry;	/* short dir entry index from start of dir */
+	unsigned int direntries;	/* number of dir entries used by this
+					 * file */
+	long dircluster;	/* first cluster of dir */
+	bool eof;
+};
+
+struct fat_dir {
+	unsigned int entry;
+	unsigned int entrycount;
+	long sector;
+	struct fat_file file;
+	unsigned char sectorcache[3][SECTOR_SIZE];
+};
+
+extern int rockbox_fat_mount(long startsector);
+extern void fat_size(unsigned long *size, unsigned long *free);	/* public for
+								 * info */
+extern void fat_recalc_free(void);	/* public for debug info screen */
+extern int fat_create_dir(const char *name,
+			  struct fat_dir *newdir, struct fat_dir *dir);
+extern long fat_startsector(void);	/* public for config sector */
+extern int fat_open(long cluster,
+		    struct fat_file *ent, const struct fat_dir *dir);
+extern int fat_create_file(const char *name,
+			   struct fat_file *ent, struct fat_dir *dir);
+extern long fat_readwrite(struct fat_file *ent, long sectorcount,
+			  void *buf, bool write);
+extern int fat_closewrite(struct fat_file *ent, long size, int attr);
+extern int fat_seek(struct fat_file *ent, unsigned long sector);
+extern int fat_remove(struct fat_file *ent);
+extern int fat_truncate(const struct fat_file *ent);
+extern int fat_rename(struct fat_file *file,
+		      struct fat_dir *dir,
+		      const unsigned char *newname, long size, int attr);
+
+extern int fat_opendir(struct fat_dir *ent, unsigned long currdir,
+		       const struct fat_dir *parent_dir);
+extern int fat_getnext(struct fat_dir *ent, struct fat_direntry *entry);
+
+#ifndef _ROCKBOX_FILE_H_
+#include <rockbox_file.h>
+#endif
+
+#ifndef _ROCKBOX_DIR_H
+#include <rockbox_dir.h>
+#endif
+
+#endif				/* #ifndef ROCKBOX_FAT_H */
diff --git a/include/rockbox_file.h b/include/rockbox_file.h
new file mode 100644
index 0000000..d910d64
--- /dev/null
+++ b/include/rockbox_file.h
@@ -0,0 +1,61 @@
+/***************************************************************************
+ *             __________               __   ___.
+ *   Open      \______   \ ____   ____ |  | _\_ |__   _______  ___
+ *   Source     |       _//  _ \_/ ___\|  |/ /| __ \ /  _ \  \/  /
+ *   Jukebox    |    |   (  <_> )  \___|    < | \_\ (  <_> > <  <
+ *   Firmware   |____|_  /\____/ \___  >__|_ \|___  /\____/__/\_ \
+ *                     \/            \/     \/    \/            \/
+ * $Id: file.h,v 1.14 2005/02/28 18:32:57 amiconn Exp $
+ *
+ * Copyright (C) 2002 by Bjrn Stenberg
+ *
+ * All files in this archive are subject to the GNU General Public License.
+ * See the file COPYING in the source tree root for full license agreement.
+ *
+ * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
+ * KIND, either express or implied.
+ *
+ * 01/17/2006 Keith Outwater (outwater4@comcast.net) - port to U-Boot using
+ *     CVS version 1.15 of 'firmware/common/file.h' from rockbox CVS server.
+ *
+ ****************************************************************************/
+
+#ifndef _ROCKBOX_FILE_H_
+#define _ROCKBOX_FILE_H_
+
+#include <linux/types.h>
+
+#define MAX_PATH 260
+
+#ifndef SEEK_SET
+#define SEEK_SET 0
+#endif
+#ifndef SEEK_CUR
+#define SEEK_CUR 1
+#endif
+#ifndef SEEK_END
+#define SEEK_END 2
+#endif
+
+#ifndef O_RDONLY
+#define O_RDONLY 0
+#define O_WRONLY 1
+#define O_RDWR   2
+#define O_CREAT  4
+#define O_APPEND 8
+#define O_TRUNC  0x10
+#endif
+
+extern int open(const char *pathname, int flags);
+extern int close(int fd);
+extern int fsync(int fd);
+extern ssize_t read(int fd, void *buf, size_t count);
+extern off_t lseek(int fildes, off_t offset, int whence);
+extern int creat(const char *pathname, mode_t mode);
+extern ssize_t write(int fd, const void *buf, size_t count);
+extern int remove(const char *pathname);
+extern int rename(const char *path, const char *newname);
+extern int ftruncate(int fd, off_t length);
+extern off_t filesize(int fd);
+
+#endif				/* #ifndef _ROCKBOX_FILE_H_ */
